from pydantic import ValidationError

from tests.integration.graph import is_edge


def test_negative_trivial_schema_load(fhir_schema, fhir_validator):
    """a negative test - all should fail"""

    for vertex_name in fhir_schema['$defs']:

        if not is_edge(fhir_schema['$defs'][vertex_name]):
            continue

        try:
            # a trivial instance example
            fhir_validator.validate(
                {"id": 1234, "resourceType": vertex_name},
            )
            raise Exception(f'!ok - {vertex_name} should have raised ValidationError')
        except ValidationError:
            pass

    print('  ok - caught invalid id')



import logging
import pathlib
import subprocess
from collections import defaultdict
from typing import List, Iterator, Any

from fastavro import reader
from pydantic import BaseModel

logger = logging.getLogger(__name__)


def run_cmd(command_line) -> str:
    """Run a command line, return stdout."""
    try:
        logger.debug(command_line)
        return subprocess.check_output(command_line, shell=True).decode("utf-8").rstrip()
    except Exception as exc:
        logger.error(exc)
        raise exc


class EdgeSummary(BaseModel):
    """Summary of edge in PFB."""

    src: str = None
    dst: str = None
    count: int = 0


class EntitySummary(BaseModel):
    """Summary of entity in PFB."""

    name: str = None
    count: int = 0
    relationships: dict[str, EdgeSummary] = {}


class InspectionResults(BaseModel):
    """Results of PFB inspection."""

    errors: List[str] = []
    warnings: List[str] = []
    info: List[str] = []
    counts: dict[str, EntitySummary] = {}


def inspect_pfb(file_name) -> InspectionResults:
    """Show details of the pfb."""
    # TODO - simplify
    results = InspectionResults()
    # raw records
    records = []
    # TODO - should this be a generator?
    with open(file_name, 'rb') as fo:
        records = [record for record in reader(fo)]

    # ensure loaded in the correct order

    def recursive_default_dict():
        """Recursive default dict."""
        return defaultdict(recursive_default_dict)

    seen_already = set()

    def log_first_occurrence(obj_):
        seen_already.add(obj_['name'])

    def check_links(obj_, graph_) -> Iterator[str]:
        """Make sure relations from obj exist in graph."""
        assert 'relations' in obj_
        for r in obj_['relations']:
            if not graph_[r['dst_name']][r['dst_id']]:
                yield f"{r['dst_name']}.{r['dst_id']} referenced from {obj_['name']}.{obj_['id']} not found in Graph"

    graph = recursive_default_dict()

    for obj in records:
        log_first_occurrence(obj)
        graph[obj['name']][obj['id']] = obj
        for error in check_links(obj, graph):
            results.errors.append(error)

    # ensure no duplicates
    seen_already = set()

    def check_duplicates(obj_):
        if obj_['id'] in seen_already:
            yield f"Duplicate {obj_['name']}/{obj_['id']}"
        seen_already.add(obj_['id'])

    for obj in records:
        for error in check_duplicates(obj):
            results.errors.append(error)

    with_relations = [r for r in records if len(r['relations']) > 0]
    results.info.append(f"'Records with relationships': {len(with_relations)}")
    results.info.append(f"'Records': {len(records)}")

    assert len(records) > 1, f"Should have more than just metadata {file_name}"
    if len(with_relations) == 0:
        results.warnings.append("No records have relationships.")

    for obj in records:
        if obj['name'] not in results.counts:
            results.counts[obj['name']] = EntitySummary(name=obj['name'])
        summary = results.counts[obj['name']]
        summary.count += 1
        for r in obj['relations']:
            if r['dst_name'] not in summary.relationships:
                summary.relationships[r['dst_name']] = EdgeSummary(src=summary.name, dst=r['dst_name'])
            edge_summary = summary.relationships[r['dst_name']]
            edge_summary.count += 1

    return results


DEFAULT_DEPENDENCY_ORDER = """
_definitions.yaml
_terms.yaml
Program
Project
ResearchStudy
Patient
ResearchSubject
Substance
Specimen
Encounter
Observation
Condition
Medication
MedicationAdministration
DocumentReference
Task
FamilyMemberHistory
BodyStructure
""".split()


class SimplePFBWriter:
    """Create a PFB."""

    def __init__(self, schema_path: str, output_path: Any,
                 dependency_order: list[str] = DEFAULT_DEPENDENCY_ORDER):
        """Add schema to pfb file.

        see https://github.com/uc-cdis/pypfb/blob/d7cbb64813a9f1b9006a1ec0f92aa86962df2709/src/pfb/importers/gen3dict.py#L15
        """
        self.schema_path = schema_path

        if isinstance(output_path, str):
            output_path = pathlib.Path(output_path)
        output_path = output_path.expanduser()
        assert output_path.parent.is_dir(), f"{output_path.parent} is not a directory"
        self.output_path = output_path

        self.schema_written = False

        self.dependency_order = dependency_order

    def write_schema(self):
        """Create pfb file with the schema."""
        cmd = f"pfb from -o {self.output_path} dict {self.schema_path}"
        logger.info(f"Creating pfb file {self.output_path}  with schema read from: {self.output_path}")
        self.schema_written = True
        # TODO - We need to shell out to do this,  _from_dict is not public
        #  https://github.com/uc-cdis/pypfb/blob/master/src/pfb/importers/gen3dict.py#L29
        return run_cmd(cmd)

    def write(self, ndjson_file: str) -> str:
        """Add data to pfb file."""

        if not self.schema_written:
            self.write_schema()

        cmd = f"pfb add -i {ndjson_file} {self.output_path}"
        logger.info(f"Adding {ndjson_file}")
        # TODO - We need to shell out to do this,  _from_dict is not public
        #  https://github.com/uc-cdis/pypfb/blob/master/src/pfb/importers/gen3dict.py#L29
        return run_cmd(cmd)

    def inspect(self) -> InspectionResults:
        """Show details of the pfb."""
        return inspect_pfb(self.output_path)

    def transform_directory(self, input_path: Any) -> Iterator[str]:
        """Write json files from directory in order."""
        if isinstance(input_path, str):
            input_path = pathlib.Path(input_path)
        assert input_path.exists() and input_path.is_dir(), f"{input_path} should be a directory"
        write_order = []
        files = [_ for _ in input_path.glob('*.ndjson')]
        for _ in self.dependency_order:
            file = next(iter([n for n in files if f"{_}.ndjson" in str(n)]), None)
            if file:
                write_order.append(file)
        for file in write_order:
            yield self.write(file)



from dictionaryutils import DataDictionary


def test_codeable_concept(data_dictionary_from_yaml: DataDictionary):
    """Schema has Gen3 friendly CodeableConcept."""
    schemas = data_dictionary_from_yaml.schema
    condition = schemas['condition']
    code = condition['properties']['code']
    code_coding = condition['properties']['code_coding']
    category = condition['properties']['category']
    category_coding = condition['properties']['category_coding']
    verification_status = condition['properties']['verificationStatus']

    assert 'text representation' in code['description']
    assert 'text representation' in category['description']
    assert 'system#code representation' in code_coding['description']
    assert 'system#code representation' in category_coding['description']
    assert verification_status['binding_strength'] == 'required'
    assert verification_status['binding_uri']



from fhir.resources.documentreference import DocumentReference
from fhir.resources.patient import Patient, PatientLink, PatientCommunication
from fhir.resources.task import Task

from iceberg_tools.schema import _find_fhir_classes, extract_schemas, BASE_URI


def test_find_fhir_classes():
    """Assert we find classes descending from Patient."""
    classes = _find_fhir_classes({'dependency_order': ['Patient']})
    expected_classes = [Patient, PatientLink, PatientCommunication]
    for _ in expected_classes:
        assert _ in classes, ('No find', _)


def test_schemas():
    """Assert we find schemas descending from Patient."""
    classes = _find_fhir_classes({'dependency_order': ['Patient']})
    schemas = extract_schemas(classes, BASE_URI)
    expected_classes = ['Patient', 'PatientLink', 'PatientCommunication']
    for _ in expected_classes:
        assert _ in schemas, ('No find expected class', _)

    patient = schemas['Patient']
    assert patient['$id'] == 'Patient'
    assert 'Disclaimer' not in patient['description'], "Did not edit vertex description"


def test_bindings():
    """Ensure bindings are attached to expected properties."""
    schemas = extract_schemas([Patient], BASE_URI)
    patient = schemas['Patient']
    gender = patient['properties']['gender']
    expected_property_attributes = ['type', 'binding_strength', 'binding_description', 'binding_uri', 'enum_values']
    for _ in expected_property_attributes:
        assert _ in gender, ('Did not find binding. Missing attribute in Patient.gender', _)


def test_backref():
    """Test simple and itemized backref."""
    schemas = extract_schemas([Task], BASE_URI)
    assert 'Task' in schemas, "Should have found Task"

    task = schemas['Task']
    expected_references_with_itemized_backref = ['focus', 'owner', 'requester']
    for _ in expected_references_with_itemized_backref:
        assert _ in task['properties'], task['properties'].keys()
        assert task['properties'][_]['backref'] == f'{_}_task', "Should have itemized backref"
    expected_references_with_simple_backref = ['location']
    for _ in expected_references_with_simple_backref:
        assert task['properties'][_]['backref'] == 'task', "Should have simple backref"


def test_backref_array():
    """Test simple and itemized backref."""
    schemas = extract_schemas([DocumentReference], BASE_URI)
    assert 'DocumentReference' in schemas, "Should have found DocumentReference"

    document_reference = schemas['DocumentReference']
    # select property that is a list of references.
    author = document_reference['properties']['author']
    print(author)
    assert 'backref' in author, "lists of References should have a backref"


def test_task():
    """Test CodeableReference."""
    schemas = extract_schemas([Task], BASE_URI)
    assert 'Task' in schemas, "Should have found Task"

    task = schemas['Task']
    # select property that is a CodeableReference.
    requested_performer = task['properties']['requestedPerformer']
    print(requested_performer)
    assert 'backref' in requested_performer, "requested_performer should have a backref"




def test_embedded_types(distribution_schema: dict):
    """Ensure that subtypes are described as simple `string`"""
    patient = distribution_schema['patient.yaml']
    # lists
    assert patient['properties']['name']['items']['type'] == 'string'
    # scalars
    condition = distribution_schema['condition.yaml']
    condition['properties']['encounter']['type'] == 'string'



from collections import defaultdict
from typing import Dict

import pathlib

import inflection
from fastavro import reader
from pathlib import Path
import logging

logger = logging.getLogger(__name__)


def recursive_default_dict() -> Dict:
    """Recursive default dict."""
    return defaultdict(recursive_default_dict)


def aggregate_edges(path, output_path, pattern):
    """Aggregate avro pfb files into a cytoscape friendly tsv."""
    aggregated_node_counts = defaultdict(dict)
    aggregated_edge_counts = recursive_default_dict()

    path = Path(path)
    for file_path in path.glob(pattern):
        with open(file_path, 'rb') as fo:
            logger.info(f"Reading {file_path}")
            avro_reader = reader(fo)
            for record in avro_reader:
                # skip metadata
                if record['name'] == 'Metadata':
                    continue

                if record['name'] not in aggregated_node_counts[file_path]:
                    aggregated_node_counts[file_path][record['name']] = 0
                aggregated_node_counts[file_path][record['name']] += 1

                for relation in record['relations']:
                    if relation['dst_name'] not in aggregated_edge_counts[record['name']]:
                        aggregated_edge_counts[record['name']][relation['dst_name']] = set()
                    aggregated_edge_counts[record['name']][relation['dst_name']].add(str(file_path))

    fn = pathlib.Path(output_path) / "aggregated_edge_counts.tsv"
    with open(fn, "w") as fp:
        print("source\ttarget\tsource_count\tedge_count", file=fp)
        for source in aggregated_edge_counts:
            for target in aggregated_edge_counts[source]:
                print(
                    f"{source}\t{target}\t{len(aggregated_edge_counts[source][target])}\t{len(aggregated_node_counts[source])}",
                    file=fp)
    logger.info(f"Wrote {fn}")

    fn = pathlib.Path(output_path) / "aggregated_node_counts.tsv"
    with open(fn, "w") as fp:
        print("file_path\tentity\tcount", file=fp)
        for file_path in aggregated_node_counts:
            for entity, count in aggregated_node_counts[file_path].items():
                print(f"{file_path}\t{entity}\t{count}", file=fp)
    logger.info(f"Wrote {fn}")

    pivot_aggregated_node_counts = defaultdict(dict)
    entity_names = set()
    file_names = set()
    for file_path in aggregated_node_counts:
        for entity, count in aggregated_node_counts[file_path].items():
            entity = inflection.camelize(entity)
            entity_names.add(entity)
            file_names.add(file_path.stem)
            pivot_aggregated_node_counts[entity][file_path.stem] = count

    entity_names = sorted(entity_names)
    file_names = sorted(file_names)
    rows = defaultdict(dict)
    for entity_name in entity_names:
        for file_name in file_names:
            if file_name not in pivot_aggregated_node_counts[entity_name]:
                rows[entity_name][file_name] = ''
            else:
                rows[entity_name][file_name] = pivot_aggregated_node_counts[entity_name][file_name]

    fn = pathlib.Path(output_path) / "pivoted_aggregated_node_counts.tsv"
    with open(fn, "w") as fp:
        headers = '\t'.join(['entity'] + file_names)
        print(headers, file=fp)
        for entity_name in rows:
            row = '\t'.join([entity_name] + [str(rows[entity_name][file_name]) for file_name in file_names])
            print(row, file=fp)
    logger.info(f"Wrote {fn}")



import subprocess


def test_lathe_lint(bmeg_dir, lath_path):
    """Run lathe's schema-lint command."""
    cmd = f"{lath_path} schema-lint .".split()
    results = subprocess.run(cmd)
    assert results.returncode == 0, (results.stdout, results.stderr)



import logging

import click

from iceberg_tools.cli.data import cli as data
from iceberg_tools.cli.schema import cli as schema
from iceberg_tools.util import NaturalOrderGroup

logger = logging.getLogger(__name__)


@click.group(cls=NaturalOrderGroup)
def cli():
    """Manage FHIR based schemas and data."""
    pass


cli.add_command(schema)
cli.add_command(data)


if __name__ == '__main__':
    cli(auto_envvar_prefix='ICEBERG')



from pydantic import ValidationError


def test_edge_load_granular_reference(fhir_schema, fhir_validator, edge_schemas):
    fail = False
    for edge in edge_schemas:
        try:
            fhir_validator.validate(
                {
                    "resourceType": edge['title'],
                    "source": {"type": "Patient", "identifier": {"value": "1234"}},
                    "target": {"type": "Specimen", "identifier": {"value": "5678"}},
                }
            )
        except ValidationError:
            fail = True
            print('fail - fhir_schema should have validated granular Reference', edge['title'])
        # print(f"ok {edge['title']}")
    if not fail:
        print('  ok - all edges validated correctly with type/identifier References')



from collections import defaultdict

from jsonschema.exceptions import ValidationError

from tests.integration.graph import edge_validator


def test_edge_validator(fhir_schema, fhir_validator, edge_schemas):

    for edge in edge_schemas:
        try:
            edge_validator(fhir_validator, {
                "resourceType": edge['title'],
                "source": {"type": "XXX", "identifier": {"value": "1234"}},
                "destination": {"type": "YYYY", "identifier": {"value": "5678"}},
            })
            assert False, ("Should not validate", edge)
        except ValidationError:
            pass
    print('  ok all invalid edges pass "extra" vocabulary validation')

    fail = False
    for edge in edge_schemas:
        instance = {
                "resourceType": edge['title'],
                "source": {"type": edge['source_type'], "identifier": {"value": "1234"}},
                "target": {"type": edge['destination_type'], "identifier": {"value": "5678"}},
            }
        try:
            edge_validator(fhir_validator, instance)
        except ValidationError:
            fail = True
            print(f"fail extra edge_validator {instance}")
    if not fail:
        print('  ok all valid edges pass "extra" vocabulary validation')

    print('count of primary edges:\n  ', len([edge for edge in edge_schemas if edge['is_primary']]))

    print('unresolved `Any` edge destinations:')
    needs_unresolved_primary = False
    for edge in edge_schemas:
        if edge['destination_type'] == 'Resource' and len([e for e in edge_schemas if e['source_type'] == edge['source_type']]) == 1:
            print('  ', edge['source_type']+'.'+edge['source_property_name'], edge['destination_type'])
            needs_unresolved_primary = True
    if not needs_unresolved_primary:
        print('  None')

    print('multi-edges (ie need `is_primary` manually set for uni-graphs (simplified)):')

    edge_counts = defaultdict(dict)
    for edge in edge_schemas:
        if edge['destination_type'] not in edge_counts[edge['source_type']]:
            edge_counts[edge['source_type']][edge['destination_type']] = False
        if not edge_counts[edge['source_type']][edge['destination_type']]:
            edge_counts[edge['source_type']][edge['destination_type']] = edge['is_primary']

    needs_is_primary = False
    for source_type in edge_counts:
        for destination_type in edge_counts[source_type]:
            if not edge_counts[source_type][destination_type]:
                if destination_type == 'Resource':
                    continue
                print(source_type, '->', destination_type)
                needs_is_primary = True

    if not needs_is_primary:
        print('  None')



# http://build.fhir.org/specimen-example.json

import orjson
from fhir.resources.patient import Patient
from fhir.resources.specimen import Specimen

from iceberg_tools.data.simplifier import simplify, _default_json_serializer, SimplifierContextManager
from iceberg_tools.schema.simplified_validator import ensure_schema, validate

SPECIMEN = {
    "resourceType": "Specimen",
    "id": "101",
    "text": {
        "status": "generated",
        "div": "TEXT ..."
    },
    "identifier": [{
        "system": "http://ehr.acme.org/identifiers/collections",
        "value": "23234352356"
    }],
    "accessionIdentifier": {
        "system": "http://lab.acme.org/specimens/2011",
        "value": "X352356"
    },
    "status": "available",
    "type": {
        "coding": [{
            "system": "http://snomed.info/sct",
            "code": "122555007",
            "display": "Venous blood specimen"
        }]
    },
    "subject": {
        "reference": "Patient/example",
        "display": "Peter Patient"
    },
    "receivedTime": "2011-03-04T07:03:00Z",
    "request": [{
        "reference": "ServiceRequest/example"
    }],
    "collection": {
        "collector": {
            "reference": "Practitioner/example"
        },
        "collectedDateTime": "2011-05-30T06:15:00Z",
        "quantity": {
            "value": 6,
            "unit": "mL"
        },
        "method": {
            "coding": [{
                "system": "http://terminology.hl7.org/CodeSystem/v2-0488",
                "code": "LNV"
            }]
        },
        "bodySite": {
            "concept": {
                "coding": [{
                    "system": "http://snomed.info/sct",
                    "code": "49852007",
                    "display": "Structure of median cubital vein (body structure)"
                }]
            }
        }
    },
    "container": [{
        "device": {
            "reference": "Device/device-example-specimen-container-green-gel-vacutainer"
        },
        "specimenQuantity": {
            "value": 3,
            "unit": "mL"
        }
    }],
    "note": [{
        "text": "Specimen is grossly lipemic"
    }],
    "processing": [{
        "description": "Acidify to pH < 3.0 with 6 N HCl.",
        "method": {
            "coding": [{
                "system": "http://terminology.hl7.org/CodeSystem/v2-0373",
                "code": "ACID"
            }]
        },
        "additive": [{
            "reference": "Substance/CHEMBL1231821",
            "display": "6 N HCl"
        }],
        "timeDateTime": "2015-08-18T08:10:00Z"
    }],
}

PATIENT = {
    "resourceType": "Patient",
    "id": "0e67ef67-bc88-4792-aa70-ccef7db25153",
    "meta": {
        "versionId": "1",
        "lastUpdated": "2023-01-26T14:55:20.239+00:00",
        "source": "#7mKbhP1TZTy3Y1vr",
        "profile": [
            "http://hl7.org/fhir/us/core/StructureDefinition/us-core-patient"
        ]
    },
    "text": {
        "status": "generated",
        "div": "<div> generated</div>"
    },
    "extension": [
        {
            "url": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-race",
            "extension": [
                {
                    "url": "ombCategory",
                    "valueCoding": {
                        "system": "urn:oid:2.16.840.1.113883.6.238",
                        "code": "2106-3",
                        "display": "White"
                    }
                },
                {
                    "url": "text",
                    "valueString": "White"
                }
            ]
        },
        {
            "url": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-ethnicity",
            "extension": [
                {
                    "url": "ombCategory",
                    "valueCoding": {
                        "system": "urn:oid:2.16.840.1.113883.6.238",
                        "code": "2186-5",
                        "display": "Non Hispanic or Latino"
                    }
                },
                {
                    "url": "text",
                    "valueString": "Non Hispanic or Latino"
                }
            ]
        },
        {
            "url": "http://hl7.org/fhir/StructureDefinition/patient-mothersMaidenName",
            "valueString": "Pandora807 Toy286"
        },
        {
            "url": "http://hl7.org/fhir/us/core/StructureDefinition/us-core-birthsex",
            "valueCode": "M"
        },
        {
            "url": "http://hl7.org/fhir/StructureDefinition/patient-birthPlace",
            "valueAddress": {
                "city": "Medford",
                "state": "Massachusetts",
                "country": "US"
            }
        },
        {
            "url": "http://synthetichealth.github.io/synthea/disability-adjusted-life-years",
            "valueDecimal": 4.840278904991756
        },
        {
            "url": "http://synthetichealth.github.io/synthea/quality-adjusted-life-years",
            "valueDecimal": 73.15972109500825
        }
    ],
    "identifier": [
        {
            "system": "https://github.com/synthetichealth/synthea",
            "value": "e79d01d9-25bd-ee33-e40b-34cb119ca8e4"
        },
        {
            "type": {
                "coding": [
                    {
                        "system": "http://terminology.hl7.org/CodeSystem/v2-0203",
                        "code": "MR",
                        "display": "Medical Record Number"
                    }
                ],
                "text": "Medical Record Number"
            },
            "system": "http://hospital.smarthealthit.org",
            "value": "e79d01d9-25bd-ee33-e40b-34cb119ca8e4"
        },
        {
            "type": {
                "coding": [
                    {
                        "system": "http://terminology.hl7.org/CodeSystem/v2-0203",
                        "code": "SS",
                        "display": "Social Security Number"
                    }
                ],
                "text": "Social Security Number"
            },
            "system": "http://hl7.org/fhir/sid/us-ssn",
            "value": "999-15-5094"
        },
        {
            "type": {
                "coding": [
                    {
                        "system": "http://terminology.hl7.org/CodeSystem/v2-0203",
                        "code": "DL",
                        "display": "Driver's License"
                    }
                ],
                "text": "Driver's License"
            },
            "system": "urn:oid:2.16.840.1.113883.4.3.25",
            "value": "S99947903"
        },
        {
            "type": {
                "coding": [
                    {
                        "system": "http://terminology.hl7.org/CodeSystem/v2-0203",
                        "code": "PPN",
                        "display": "Passport Number"
                    }
                ],
                "text": "Passport Number"
            },
            "system": "http://standardhealthrecord.org/fhir/StructureDefinition/passportNumber",
            "value": "X47675657X"
        }
    ],
    "name": [
        {
            "use": "official",
            "family": "Walsh511",
            "given": [
                "Nestor901"
            ],
            "prefix": [
                "Mr."
            ]
        }
    ],
    "telecom": [
        {
            "system": "phone",
            "value": "555-905-1070",
            "use": "home"
        }
    ],
    "gender": "male",
    "birthDate": "1917-05-31",
    "deceasedDateTime": "1996-12-19T02:00:49-05:00",
    "address": [
        {
            "extension": [
                {
                    "url": "http://hl7.org/fhir/StructureDefinition/geolocation",
                    "extension": [
                        {
                            "url": "latitude",
                            "valueDecimal": 42.08152268673541
                        },
                        {
                            "url": "longitude",
                            "valueDecimal": -71.29566704174853
                        }
                    ]
                }
            ],
            "line": [
                "640 Bailey Spur"
            ],
            "city": "Norfolk",
            "state": "MA",
            "country": "US"
        }
    ],
    "maritalStatus": {
        "coding": [
            {
                "system": "http://terminology.hl7.org/CodeSystem/v3-MaritalStatus",
                "code": "M",
                "display": "M"
            }
        ],
        "text": "M"
    },
    "multipleBirthBoolean": False,
    "communication": [
        {
            "language": {
                "coding": [
                    {
                        "system": "urn:ietf:bcp:47",
                        "code": "en-US",
                        "display": "English"
                    }
                ],
                "text": "English"
            }
        }
    ]
}


def test_simplify_specimen():
    nested_objects = {
        'Specimen': ['processing.additive', 'processing.method', 'collection.method', 'collection.bodySite'],
        'DocumentReference': ['content.contentType', 'content.md5', 'content.size', 'content.url'],
        'Patient': ['address.postalCode']
    }
    specimen = Specimen(**SPECIMEN)
    with SimplifierContextManager():
        simplified, references = simplify(specimen, 'PFB', nested_objects)

    print(orjson.dumps(simplified, default=_default_json_serializer,
                       option=orjson.OPT_APPEND_NEWLINE).decode())

    assert simplified['collection_method_coding'] == "http://terminology.hl7.org/CodeSystem/v2-0488#LNV"
    assert simplified['collection_method'] == "LNV"

    assert simplified['collection_bodySite'] == "Structure of median cubital vein (body structure)"
    assert simplified['collection_bodySite_coding'] == "http://snomed.info/sct#49852007"

    assert simplified['processing_additive'] == ['Substance/CHEMBL1231821']

    assert simplified['processing_method'] == 'ACID'
    assert simplified['processing_method_coding'] == 'http://terminology.hl7.org/CodeSystem/v2-0373#ACID'

    assert 'Patient/example' in references
    assert 'Substance/CHEMBL1231821' in references

    schema = ensure_schema('iceberg/schemas/simplified/simplified-fhir.json')
    simplified = orjson.loads(orjson.dumps(simplified, default=_default_json_serializer))
    parse_result = validate({'object': simplified, "relations": [], 'id': simplified['id']}, schema)
    assert not parse_result.exception, parse_result.exception


def test_simplify_patient():
    patient = Patient(**PATIENT)
    with SimplifierContextManager():
        simplified, references = simplify(patient, 'PFB')
        simplified = orjson.loads(orjson.dumps(simplified, default=_default_json_serializer))
        schema = ensure_schema('iceberg/schemas/simplified/simplified-fhir.json')
        parse_result = validate({'object': simplified, "relations": [], 'id': simplified['id']}, schema)
        assert not parse_result.exception,  parse_result.exception



from jsonschema.exceptions import ValidationError


def test_invalid_edges(fhir_schema, fhir_validator, edge_schemas):
    invalid_edges = [
        # missing resourceType
        {
            "source": {"type": "Patient", "identifier": {"value": "1234"}},
            "destination": {"type": "Specimen", "identifier": {"value": "5678"}},
        },
        # missing source
        {
            "resourceType": 'Observation_performer_Patient',
            "destination": {"type": "Specimen", "identifier": {"value": "5678"}},
        },
        # missing destination
        {
            "resourceType": 'Observation_performer_Patient',
            "source": {"type": "Patient", "identifier": {"value": "1234"}},
        },
        # invalid source reference
        {
            "resourceType": 'Observation_performer_Patient',
            "source": {"foo"},
            "destination": {"type": "Specimen", "identifier": {"value": "5678"}},
        },
        # invalid destination reference
        {
            "resourceType": 'Observation_performer_Patient',
            "source": {"type": "Patient", "identifier": {"value": "1234"}},
            "destination": {"foo"},
        },
    ]

    for edge in invalid_edges:
        try:
            fhir_validator.validate(edge)
            assert False, ("Should not validate", edge)
        except ValidationError:
            pass

    print('  ok - all invalid edges validated correctly')



from jsonschema.exceptions import ValidationError
from jsonschema.validators import Draft202012Validator, validate


def test_spot_check_embedded_type(fhir_schema, fhir_validator):
    """spot check an embedded type"""
    try:
        instance = {"id": "1234", "resourceType": "HumanName", "period": {"start": "1234"}}
        assert not validate(instance, fhir_schema, format_checker=Draft202012Validator.FORMAT_CHECKER)
        print('  fail - fhir_schema period nonsense period invalid')
    except ValidationError as e:
        assert '1234' in e.message
        print('  ok - fhir_schema period nonsense period invalid correctly')

    try:
        instance = {"id": "1234", "resourceType": 'HumanName',
                    "period": {"start": "2023-01-01T00:00:00Z", "end": "2024-01-01T00:00:00Z"}}
        validate(instance, fhir_schema, format_checker=Draft202012Validator.FORMAT_CHECKER)
        # print('ok - fhir_schema period valid period validated correctly')
    except ValidationError:
        print('  fail - fhir_schema valid period should have validated ', instance)

    print('  ok - all vertices validated correctly')



from jsonschema.exceptions import ValidationError, UnknownType

from tests.integration.graph import is_edge


def test_trivial_schema_load(fhir_schema, fhir_validator):
    """Ensure whole schema can load"""
    for vertex_name in fhir_schema['$defs']:

        if not is_edge(fhir_schema['$defs'][vertex_name]):
            continue

        try:
            # a trivial instance example
            fhir_validator.validate(
                {"id": "1234", "resourceType": vertex_name},
            )
            # print(f'ok - {vertex_name} validated correctly')
        except ValidationError:
            # print(f'ok, expected error - {vertex_name}  {[c.message for c in e.context if c.schema["title"] == vertex_name]}')
            pass
        except UnknownType as e:
            print("!ok Houston we have a problem.  All types in $defs should have a schema.")
            raise e



import logging
from typing import Union

logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)
logger = logging.getLogger(__name__)

def monkey_patch_url_validate():
    # monkey patch to allow file: urls
    import fhir.resources.fhirtypes
    from pydantic import FileUrl

    original_url_validate = fhir.resources.fhirtypes.Url.validate


    @classmethod
    def better_url_validate(  # type: ignore
            cls, value: str, field: "ModelField", config: "BaseConfig"
    ) -> Union["AnyUrl", str]:
        """Allow file: urls. see https://github.com/pydantic/pydantic/issues/1983
        bugfix: addresses issue introduced with `fhir.resources`==7.0.1
        """
        if value.startswith("file:"):
            return FileUrl.validate(value, field, config)
        value = original_url_validate(value, field, config)
        return value


    fhir.resources.fhirtypes.Url.validate = better_url_validate
    logger.debug("monkey patched Url.validate")

monkey_patch_url_validate()



import json
import logging
import pathlib

import click
import yaml
from yaml import SafeLoader

from iceberg_tools.schema import _find_fhir_classes, BASE_URI, extract_schemas
from iceberg_tools.schema.graph import bundle_schemas
from iceberg_tools.schema.simplified import _simplify_schemas
from iceberg_tools.util import NaturalOrderGroup

logger = logging.getLogger(__name__)


@click.group('schema', cls=NaturalOrderGroup)
def cli():
    """Manage graph or simplified schemas from FHIR resources."""
    pass


@cli.group('generate')
def generate():
    """Generate from FHIR resources."""
    pass


@generate.command('graph')
@click.option('--output_path', required=True,
              default='iceberg/schemas/graph',
              show_default=True,
              help='Path to generated schema files'
              )
@click.option('--config_path',
              default='config.yaml',
              show_default=True,
              help='Path to config file.')
@click.option('--stats/--no-stats', default=False, is_flag=True, show_default=True,
              help="Log statistics about the FHIR classes found.")
def generate_bmeg(output_path, config_path, stats):
    """Create BMEG schemas."""

    output_path = pathlib.Path(output_path)
    output_path.mkdir(parents=True, exist_ok=True)
    config_path = pathlib.Path(config_path)
    assert output_path.is_dir()
    assert config_path.is_file()
    with open(config_path) as fp:
        gen3_config = yaml.load(fp, SafeLoader)

    classes = _find_fhir_classes(gen3_config, log_stats=stats)
    schemas = extract_schemas(classes, BASE_URI)

    output_path.mkdir(parents=True, exist_ok=True)

    for klass, schema in schemas.items():
        with open(output_path / pathlib.Path(klass + ".yaml"), "w") as fp:
            yaml.dump(schema, fp)
    logger.info(f"Individual yaml schemas written to {output_path}/*.yaml")

    path = bundle_schemas(output_path, BASE_URI)
    logger.info(f"Aggregated json schema written to {path}")


@generate.command('simplified')
@click.option('--output_path', required=True,
              default='iceberg/schemas/simplified',
              show_default=True,
              help='Path to generated schema files'
              )
@click.option('--config_path',
              default='config.yaml',
              show_default=True,
              help='Path to config file.')
@click.option('--gen3_fixtures',
              default='resources/static_gen3_fixtures',
              show_default=True,
              help='Path to gen3 static data dictionary files.')
@click.option('--stats/--no-stats', default=False, is_flag=True, show_default=True,
              help="Log statistics about the FHIR classes found.")
def generate_simplified(output_path, config_path, gen3_fixtures, stats):
    """Create simplified schemas."""

    output_path = pathlib.Path(output_path)
    output_path.mkdir(parents=True, exist_ok=True)
    config_path = pathlib.Path(config_path)
    gen3_fixtures = pathlib.Path(gen3_fixtures)
    assert gen3_fixtures.is_dir()
    assert output_path.is_dir()
    assert config_path.is_file()
    with open(config_path) as fp:
        gen3_config = yaml.load(fp, SafeLoader)

    classes = _find_fhir_classes(gen3_config, log_stats=stats)
    schemas = extract_schemas(classes, BASE_URI)
    schemas = _simplify_schemas(gen3_config, gen3_fixtures, schemas, log_stats=stats)

    # also write out yaml files
    for k, v in schemas.items():

        fn = k
        if not k.startswith('_') and 'id' in v:
            fn = v['id']
        if "yaml" not in fn:
            fn += ".yaml"

        with open(output_path / pathlib.Path(fn), "w") as fp:
            v['$schema'] = "http://json-schema.org/draft-04/schema#"
            yaml.dump(v, fp)

    logger.info(f"Simplified individual yaml schemas written to {output_path}/*.yaml")

    compile_gen3_schemas(output_path)

    logger.info(f'Simplified schema written to {output_path / pathlib.Path("simplified-fhir.json")}')


def compile_gen3_schemas(output_path):
    """create uber schema of all yaml"""
    from dictionaryutils import dump_schemas_from_dir
    with open(output_path / pathlib.Path("simplified-fhir.json"), "w") as fp:
        json.dump(dump_schemas_from_dir(output_path), fp, indent=2, sort_keys=False)


@cli.group('compile')
def compile_():
    """Create aggregated json file from individual yaml schemas"""
    pass


@compile_.command('simplified')
@click.option('--output_path', required=True,
              default='iceberg/schemas/simplified',
              show_default=True,
              help='Path to generated schema files'
              )
def compile_simplified(output_path):
    """Compile simplified schemas."""
    output_path = pathlib.Path(output_path)
    compile_gen3_schemas(output_path)
    logger.info(f'Simplified schemas written to {output_path / pathlib.Path("simplified-fhir.json")}')



import pathlib

from iceberg_tools.data.pfb import SimplePFBWriter
from iceberg_tools.data.simplifier import simplify_directory


def setup_module(module):
    """Setup the test module"""
    msg = '\n'.join(
        [
            "# please run the following before running this test",
            "iceberg schema generate simplified --config_path tests/integration/data/config.yaml --output_path tests/integration/data/schemas",
            "iceberg schema compile simplified --output_path tests/integration/data/schemas"
        ]
    )
    assert pathlib.Path('tests/integration/data/schemas/simplified-fhir.json').exists(), msg


def test_studies(caplog, dependency_order):
    """Ensure we can create a pfb file from a synthetic study."""
    simplify_directory('tests/fixtures/simplify/study/', '**/*.*', 'tmp/study/extractions',
                       'tests/integration/data/schemas/simplified-fhir.json', 'PFB', 'tests/integration/data/config.yaml')

    pfb_writer = SimplePFBWriter(schema_path='tests/integration/data/schemas/simplified-fhir.json',
                                 output_path='tmp/CohesiveDataSet.avro',
                                 dependency_order=dependency_order)

    list(pfb_writer.transform_directory('tmp/study/extractions'))

    inspection = pfb_writer.inspect()
    assert len(inspection.errors) == 0, "Unexpected errors"
    if len(inspection.warnings) > 0:
        print(inspection.warnings)


def test_synthea(caplog, dependency_order):
    """Ensure we can create a pfb file from a synthetic study."""
    simplify_directory('tests/fixtures/simplify/synthea', '**/*.*', 'tmp/synthea/extractions',
                       'tests/integration/data/schemas/simplified-fhir.json', 'PFB', 'tests/integration/data/config.yaml')

    pfb_writer = SimplePFBWriter(schema_path='tests/integration/data/schemas/simplified-fhir.json',
                                 output_path='tmp/Synthea.avro',
                                 dependency_order=dependency_order)

    list(pfb_writer.transform_directory('tmp/synthea/extractions'))

    inspection = pfb_writer.inspect()
    assert len(inspection.errors) == 0, "Unexpected errors"
    if len(inspection.warnings) > 0:
        print(inspection.warnings)


def test_kf(caplog, dependency_order):
    """Ensure we can create a pfb file from a Kids first study."""
    simplify_directory('tests/fixtures/simplify/kf', '**/*.*', 'tmp/kf/extractions',
                       'tests/integration/data/schemas/simplified-fhir.json', 'PFB', 'tests/integration/data/config.yaml')

    pfb_writer = SimplePFBWriter(schema_path='tests/integration/data/schemas/simplified-fhir.json',
                                 output_path='tmp/KidsFirst.avro',
                                 dependency_order=dependency_order)

    list(pfb_writer.transform_directory('tmp/kf/extractions'))

    inspection = pfb_writer.inspect()
    assert len(inspection.errors) == 0, "Unexpected errors"
    if len(inspection.warnings) > 0:
        print(inspection.warnings)


def test_ncpi(caplog, dependency_order):
    """Ensure we can create a pfb file from a NCPI IG examples."""

    simplify_directory('tests/fixtures/simplify/ncpi/examples-5.0', '*.*', 'tmp/ncpi/extractions',
                       'tests/integration/data/schemas/simplified-fhir.json', 'PFB', 'tests/integration/data/config.yaml')

    pfb_writer = SimplePFBWriter(schema_path='tests/integration/data/schemas/simplified-fhir.json',
                                 output_path='tmp/NCPI.avro',
                                 dependency_order=dependency_order)

    list(pfb_writer.transform_directory('tmp/ncpi/extractions'))

    inspection = pfb_writer.inspect()
    assert len(inspection.errors) == 0, "Unexpected errors"
    if len(inspection.warnings) > 0:
        print(inspection.warnings)


def test_genomics_reporting(caplog, dependency_order):
    """Ensure we can create a pfb file from a Genomics Reporting IG examples."""

    simplify_directory('tests/fixtures/simplify/genomics-reporting/examples-5.0',
                       '*.*', 'tmp/genomics-reporting/extractions',
                       'tests/integration/data/schemas/simplified-fhir.json', 'PFB', 'tests/integration/data/config.yaml')

    pfb_writer = SimplePFBWriter(schema_path='tests/integration/data/schemas/simplified-fhir.json',
                                 output_path='tmp/GenomicsReporting.avro',
                                 dependency_order=dependency_order)

    list(pfb_writer.transform_directory('tmp/genomics-reporting/extractions'))

    inspection = pfb_writer.inspect()
    assert len(inspection.errors) == 0, "Unexpected errors"
    if len(inspection.warnings) > 0:
        print(inspection.warnings)


def test_dbgap(caplog, dependency_order):
    """Ensure we can create a pfb file from dbGAP examples."""

    simplify_directory('tests/fixtures/simplify/dbgap/examples-5.0', '*.*', 'tmp/dbgap/extractions',
                       'tests/integration/data/schemas/simplified-fhir.json', 'PFB', 'tests/integration/data/config.yaml')

    pfb_writer = SimplePFBWriter(schema_path='tests/integration/data/schemas/simplified-fhir.json',
                                 output_path='tmp/dbGap.avro',
                                 dependency_order=dependency_order)

    list(pfb_writer.transform_directory('tmp/dbgap/extractions'))

    inspection = pfb_writer.inspect()
    assert len(inspection.errors) == 0, "Unexpected errors"
    if len(inspection.warnings) > 0:
        print(inspection.warnings)


def test_anvil(caplog, dependency_order):
    """Ensure we can create a pfb file from AnVIL examples."""

    simplify_directory('tests/fixtures/simplify/anvil/fhir-5.0/', '**/*.*', 'tmp/anvil/extractions',
                       'tests/integration/data/schemas/simplified-fhir.json', 'PFB', 'tests/integration/data/config.yaml')

    pfb_writer = SimplePFBWriter(schema_path='tests/integration/data/schemas/simplified-fhir.json',
                                 output_path='tmp/AnVIL.avro',
                                 dependency_order=dependency_order)

    list(pfb_writer.transform_directory('tmp/anvil/extractions'))

    inspection = pfb_writer.inspect()
    assert len(inspection.errors) == 0, "Unexpected errors"
    if len(inspection.warnings) > 0:
        print(inspection.warnings)



from typing import List

import pytest
import yaml
from yaml import SafeLoader


@pytest.fixture
def dependency_order() -> List[str]:
    """Ordered list of expected classes."""
    with open('tests/integration/data/config.yaml') as fp:
        gen3_config = yaml.load(fp, SafeLoader)
    return gen3_config['dependency_order']



import importlib
import json
from string import Formatter
from typing import List, Iterator

import fastjsonschema
import jsonschema
import pyjq
import requests
import re
from jsonpointer import resolve_pointer

from iceberg_tools.schema import extract_schemas, BASE_URI

NESTED_OBJECTS = pyjq.compile(
    '( paths | select(.[-1] == "$ref") ) as $p |   if getpath($p) != "Reference.yaml" then ( $p |  [.[] | tostring]  | join(".") ) else empty end')

NESTED_OBJECTS_IGNORE = ['Identifier', 'Extension']

REFERENCES = pyjq.compile(
    '( paths | select(.[-1] == "$ref")) as $p |   if getpath($p) == "Reference.yaml" then ( $p | join(".") ) else empty end')


def _extract_target_hints(schema_link):
    """Retrieve attributes from targetHints, using defaults."""
    target_hints = schema_link.get('targetHints', {})
    multiplicity = next(iter(target_hints.get('multiplicity', [])), 'has_many')
    directionality = next(iter(target_hints.get('directionality', [])), 'one')
    association = next(iter(target_hints.get('association', [])), False)
    return directionality, multiplicity, association


def _generate_links_from_fhir_references(schema) -> List[dict]:
    """Generate links for a schema.

    Returns:
        tuple: (links, nested_links)
    """

    # Direct links from {schema['title']}"
    links = []
    links.extend(_extract_links(schema))

    # Nested links
    nested_links = []
    for nested_schema, path in _extract_nested_schemas(schema):
        if nested_schema['title'] in NESTED_OBJECTS_IGNORE:
            continue
        extracted_links = _extract_links(nested_schema)
        if len(extracted_links) == 0:
            continue

        path_parts = path.split('.')
        parent_path = '.'.join(path_parts[1:-1]).replace('.items', '.-')

        for _ in extracted_links:
            _['$comment'] = f"From {nested_schema['title']}/{parent_path.replace('.-','')}"
            _['templatePointers'] = {"id": f"/{parent_path.replace('.', '/')}{_['templatePointers'][tp]}" for tp in _['templatePointers']}
            _['rel'] = f"{parent_path.replace('.-','')}_{_['rel']}"
        nested_links.extend(extracted_links)

    return links, nested_links


def _extract_nested_schemas(schema) -> Iterator[tuple[dict, str]]:
    """Extract $ref from elements in schema that are not `References`.

    Returns: (sub_schema, path) a tuple of the sub schema and the path to it in the passed schema
    """
    # print(json.dumps(schema))
    matches = sorted(NESTED_OBJECTS.all(schema))

    mod = importlib.import_module('fhir.resources')
    for match in matches:
        if match.startswith('properties.links') or match.startswith('links'):
            continue
        property_name = match.split('.')[1]
        if property_name.startswith('_'):
            continue

        property_ = schema['properties'][property_name]

        multiplicity = 'has_one'
        if 'items' in property_:
            multiplicity = 'has_many'

        if multiplicity == 'has_many':
            target = property_['items']['$ref'].split('.')[0]
        else:
            assert '$ref' in property_, ("TODO no $ref?", property_name, property_, match)
            target = property_['$ref'].split('.')[0]

        target_class = mod.get_fhir_model_class(target)
        sub_schemas = extract_schemas([target_class], BASE_URI)
        sub_schema = sub_schemas[target]

        yield sub_schema, match


def _extract_links(schema: dict) -> List[dict]:
    """Extract Link Description Object (LDO) from a schema.

    see https://json-schema.org/draft/2019-09/json-schema-hypermedia.html#rfc.section.6
    """
    matches = sorted(REFERENCES.all(schema))
    links = []
    for match in matches:
        direction = 'outbound'
        multiplicity = 'has_one'
        if 'items' in match:
            multiplicity = 'has_many'
        property_name = match.split('.')[1]
        property_ = schema['properties'][property_name]
        if 'enum_reference_types' not in property_:
            property_['enum_reference_types'] = ['__ANY__']
        append_postscript = len(property_['enum_reference_types']) > 1
        _path = '.'.join(match.split('.')[1:-1])
        _path = _path + '.reference'
        _path = _path.replace('.items', '.-')

        for enum_reference_type in property_['enum_reference_types']:
            rel = f"{property_name}"
            if append_postscript:
                rel += f"_{enum_reference_type}"
            links.append(
                {
                    "rel": rel,
                    "href": f"{enum_reference_type}/{{id}}",
                    "templateRequired": ["id"],
                    "templateHints": {
                        "multiplicity": [multiplicity],
                        "direction": [direction],
                        'backref': [property_['backref']],
                    },
                    "templatePointers": {
                        'id': "/" + _path.replace('.', '/')
                    },
                    'targetSchema': {'$ref': enum_reference_type},
                }
            )
    return links


def _load_schema(schema) -> dict:
    """Load JSON schema from dict or string."""
    if isinstance(schema, str):
        schema_ = requests.get(schema).json()
    elif isinstance(schema, dict):
        schema_ = schema
    else:
        raise ValueError("Schema must be a URL or a dict")
    return schema_


class AssociationSchema:
    """A JSON schema for an association."""

    def __init__(self, schema):
        """Load and compile a JSON schema."""
        if isinstance(schema, str):
            self.schema = requests.get(schema).json()
        elif isinstance(schema, dict):
            self.schema = schema
        else:
            raise ValueError("Schema must be a URL or a dict")
        self.compiled_schema = fastjsonschema.compile(self.schema)


    def validate(self, instance: dict) -> dict:
        """Validate data against schema.

        Args:
            instance: instance of data to validate
        Returns:
            dict: potentially modified instance
        """
        return self.compiled_schema(instance)

    @classmethod
    def is_association(cls, schema: dict) -> bool:
        """Is this schema an association?"""
        if 'links' not in schema:
            return False
        for _ in schema['links']:
            _, _, association = _extract_target_hints(_)
            if association:
                return True
        return False

    @classmethod
    def validate_schema_conventions(cls, schema: dict) -> bool:
        """Validate schema."""
        assert 'links' in schema, 'Schema should have links'
        assert len(schema['links']) >= 2, "Association schema should have at least two links"
        assert cls.is_association(schema), "Schema links should be an association"
        assert 'title' in schema, "Schema should have a title"
        return True


    def validate_links(self, instance: dict):
        """Validate instance.links against schema.links"""

        # verify that all links in instance are in schema
        schema_relationships = [_['rel'] for _ in self.schema['links']]
        for link in instance['links']:
            assert link['rel'] in schema_relationships, f"Instance of this association should have links to {schema_relationships}"

        return instance


class AssociationInstance:

    def __init__(self, association_schema: AssociationSchema, instance: dict = None, vertex_a: dict = None, vertex_b: dict = None):
        """Operate on json document in the context of an association schema.

        Parameters:
            association_schema: an AssociationSchema object
            instance: an instance of a JSON document that should conform to the association schema
            vertex_a: a vertex that should be associated
            vertex_b: a vertex that should be associated
        """
        self.instance = instance
        self.association_schema = association_schema
        self.regexp_cache = {}
        self.href_keys_cache = {}
        if instance is None:
            assert vertex_a is not None and vertex_b is not None, "Must provide instance or vertex_a and vertex_b"
            self.instance = self._create_instance(vertex_a, vertex_b)

    def __repr__(self):
        """Return a string representation of the instance links of the form V(id).backref<--Title-->V(id).backref"""
        parts = self.edge_parts()
        label = parts['label']

        # formulate a representation of the edge's vertices
        rels = [_ for _ in parts if _ != 'label']
        msgs = []
        for rel in rels:
            _ = parts[rel]
            msgs.append(f"{_['targetSchema']}({_[_['id_name']]}).{rel}")

        return f"{msgs[0]}<-{label}->{msgs[1]}"

    def _create_instance(self, vertex_a, vertex_b):
        """Create an instance of an association from two vertices."""
        links = []
        for schema_link, vertex in list(zip(self.association_schema.schema['links'], [vertex_a, vertex_b])):
            keys = self._extract_href_keys(schema_link['href'])
            values = [resolve_pointer(vertex, _) for _ in schema_link['templatePointers'].values()]
            links.append(
                {
                    'rel': schema_link['rel'],
                    'href': schema_link['href'].format(**dict(zip(keys, values))),
                }
            )
        return {
            'links': links
        }

    def _extract_href_keys(self, href):
        """Extract the keys from the href template, cache the result."""
        keys = self.href_keys_cache.get(href, None)
        if not keys:
            keys = [_[1] for _ in Formatter().parse(href)]
            self.href_keys_cache[href] = keys
        return keys

    def _extract_id(self, schema_href, instance_href) -> str:
        """Read the id from the href template."""
        rexp = self.regexp_cache.get(schema_href, None)
        if rexp is None:
            _ = schema_href
            keys = self._extract_href_keys(schema_href)
            for key in keys:
                _ = _.replace(f"{{{key}}}", '(.*)')
            rexp = re.compile(_)
            self.regexp_cache[schema_href] = rexp
        m = rexp.match(instance_href)
        assert m is not None, f"Unable to find id in instance link {instance_href} given schema {schema_href}"
        return m.group(1), self._extract_href_keys(schema_href)[0]

    def edge_parts(self):
        """Decompose the edge into its parts."""

        title = self.association_schema.schema['title']
        parts = {'label': title}

        for link in self.association_schema.schema['links']:
            _ = self._extract_link_parts(link, self.instance)
            parts[_['rel']] = _

        return parts

    def _extract_link_parts(self, schema_link: dict, instance: dict) -> dict:
        """Extract the parts of a link from the schema and instance."""
        instance_link = next(iter([_ for _ in instance['links'] if _['rel'] == schema_link['rel']]), None)

        id_, id_name = self._extract_id(schema_link['href'], instance_link['href'])

        directionality, multiplicity, _ = _extract_target_hints(schema_link)

        return {
            id_name: id_,
            'rel': schema_link['rel'],
            'targetSchema': schema_link['targetSchema']['$ref'],
            'multiplicity': multiplicity,
            'directionality': directionality,
            'id_name': id_name,
        }


class VertexSchemaDecorator:
    """Adds links to vertex schema."""

    def __init__(self, schema: dict):
        """Load and compile a JSON schema."""
        self.schema = _load_schema(schema)
        # add links property
        self.schema['properties']['links'] = {
            'links': {
                'type': 'array',
                'items': {
                    '$ref': 'https://json-schema.org/draft/2020-12/links'
                }
            }
        }
        # add links element
        links, nested_links = _generate_links_from_fhir_references(schema)
        self.schema['links'] = links + nested_links
        # check schema
        jsonschema.Draft202012Validator.check_schema(schema)


def cast_json_pointer_to_jq(_):
    """Convert a JSON pointer to a jq query.
    turn the pointer into a jq expression, note the leading dot and `-` for lists"""
    _ = '.' + _.replace('/', '.').replace('.-', '.[]?').replace('.', ' | .')
    _ = re.sub(r'.(\d+)', r'.[\g<1>]', _)
    return _


class VertexLinkWriter:
    """Context manager for inserting links into vertices."""

    def __init__(self, schema: dict):
        """Initialize the context manager.

        Parameters:
            schema: a schema
        """
        self.vertex_schema = schema
        self.href_keys_cache = {}
        self.jq_cache = {}
        self.regexp_cache = {}

    def _extract_href_keys(self, href):
        """Extract the keys from the href template, cache the result."""
        keys = self.href_keys_cache.get(href, None)
        if not keys:
            keys = [_[1] for _ in Formatter().parse(href)]
            self.href_keys_cache[href] = keys
        return keys

    def __enter__(self):
        """Enter the context."""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Exit the context."""
        pass

    def insert_links(self, vertex: dict) -> dict:
        """Insert links into a vertex.

        Parameters:
            vertex: a vertex
        Returns:
            dict: vertex with links inserted

        """
        links = []

        _schema = self.vertex_schema
        if isinstance(self.vertex_schema, VertexSchemaDecorator):
            _schema = self.vertex_schema.schema

        for schema_link in _schema['links']:

            keys = self._extract_href_keys(schema_link['href'])

            values = self._extract_values(schema_link, vertex)

            if None in values:
                continue

            for _ in values:
                if not isinstance(_, list):
                    _ = [_]
                links.append(
                    {
                        'rel': schema_link['rel'],
                        'href': schema_link['href'].format(**dict(zip(keys, _))),
                    }
                )

        vertex['links'] = links
        return vertex

    def validate(self, vertex: dict):
        """Validate a vertex.

        Parameters:
            vertex: a vertex
        Returns:
            jsonschema.ValidationError if the vertex is invalid, True otherwise

        """
        jsonschema.validate(vertex, self.schema)
        return True

    def _extract_values(self, schema_link, vertex) -> list:
        """Extract values from a vertex given a link description object"""
        values = []
        for k, v in schema_link['templatePointers'].items():
            values_ = self.extract_json_pointer_via_jq(v, vertex)

            if None in values_:
                #  Unable to resolve {schema_link['href']} {schema_link['templatePointers']}
                return [None]

            # disambiguate polymorphic references
            # if we have a FHIR polymorphic reference, we need to check that the target schema is in the list of values
            _ = "".join(values_)
            if '/' in _ and schema_link['targetSchema']['$ref'] not in _:
                # Polymorphic reference {schema_link['targetSchema']['$ref']} {values} skipping
                return [None]

            # Resolved {schema_link['href']} {schema_link['templatePointers']} {values_}
            # strip the prefix for fhir references
            values.extend([self._extract_value(schema_link['href'], _)[0] for _ in values_])
        return values

    def extract_json_pointer_via_jq(self, json_pointer, vertex):
        """Convert json pointer to jq query, cache the compiled and extract the value from the vertex."""
        if json_pointer not in self.jq_cache:
            jq = cast_json_pointer_to_jq(json_pointer)
            self.jq_cache[json_pointer] = self.jq_cache.get(json_pointer, None) or pyjq.compile(jq)
        values_ = self.jq_cache[json_pointer].all(vertex)
        return values_

    def _extract_value(self, schema_href, instance_href) -> (str, str):
        """Read the value from instance href, strips fhir relative references.
        Returns: (instance_value, key)
            """
        # not a fhir relative reference
        if '/' not in instance_href:
            key = self._extract_href_keys(schema_href)[0]
            return instance_href, key

        # cache the regexp
        rexp = self.regexp_cache.get(schema_href, None)
        if rexp is None:
            _ = schema_href
            keys = self._extract_href_keys(schema_href)
            for key in keys:
                _ = _.replace(f"{{{key}}}", '(.*)')
            rexp = re.compile(_)
            self.regexp_cache[schema_href] = rexp
        m = rexp.match(instance_href)
        assert m is not None, f"Unable to find value in instance link {instance_href} given schema {schema_href}"
        _ = self._extract_href_keys(schema_href)[0]
        return m.group(1), _


class SchemaLinkWriter:
    """Context manager for inserting links into schemas."""

    def __init__(self):
        """Initialize the context manager.
        """
        pass

    def __enter__(self):
        """Enter the context."""
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Exit the context."""
        pass

    @staticmethod
    def insert_links(schema) -> dict:
        """Insert links into a schema.

        Parameters:
            schema: a schema dict or url
        Returns:
            dict: schema with links inserted

        """
        schema = _load_schema(schema)
        links, nested_links = _generate_links_from_fhir_references(schema)
        schema['links'] = links + nested_links
        schema['properties']['links'] = {
            'links': {
                'type': 'array',
                'items': {
                    '$ref': 'https://json-schema.org/draft/2020-12/links'
                }
            }
        }
        return schema

    @staticmethod
    def validate(schema: dict):
        """Validate a schema.

        Parameters:
            schema: a schema
        Returns:
            jsonschema.ValidationError if the schema is invalid, True otherwise

        """
        jsonschema.Draft202012Validator.check_schema(schema)
        return True



import copy
import importlib
import logging
import pathlib
import sqlite3
from collections import defaultdict
from typing import List

import inflection
import orjson
from fhir.resources.core.utils.common import get_fhir_type_name
from fhir.resources.fhirprimitiveextension import FHIRPrimitiveExtension

from iceberg_tools.schema.fhir_resources.binding_lookup import create_fhir_definitions_lookup


logger = logging.getLogger(__name__)

path = pathlib.Path("~/.iceberg").expanduser()
path.mkdir(parents=True, exist_ok=True)
path = path / "element_bindings.sqlite"
if not path.is_file():
    create_fhir_definitions_lookup()
if not path.is_file():
    logger.error(f"Unable to find fhir bindings at {path}")
    exit(2)

ELEMENT_DB = sqlite3.connect(path)
BASE_URI = 'http://graph-fhir.io/schema/0.0.2'


def _find_fhir_classes(gen3_config, log_stats=True) -> List[type]:
    """Based on config, expand dependencies."""

    class_names = [c for c in gen3_config['dependency_order'] if not c.startswith('_')]
    class_names = [c for c in class_names if c not in ['Program', 'Project']]

    subclass_depth = gen3_config.get('subclass_depth', 3)

    mod = importlib.import_module('fhir.resources')
    classes = set()
    for class_name in class_names:
        classes.add(mod.get_fhir_model_class(class_name))

    # find subclasses N levels deep
    # maintain statistics
    stats = defaultdict(int)

    for _ in range(subclass_depth):
        embedded_classes = set()
        for klass in classes:

            for p in klass.element_properties():
                try:

                    embedded_class = mod.get_fhir_model_class(get_fhir_type_name(p.type_))
                    embedded_classes.add(embedded_class)

                    _update_stats(stats, klass, embedded_class)

                except KeyError:
                    pass

        classes.update(embedded_classes)
    classes.add(FHIRPrimitiveExtension)

    if log_stats:
        logger.info(f"Class statistics number of nested objects:\n{_summarize_stats(stats)}")

    return classes


def _summarize_stats(stats: dict) -> str:
    """Summarize stats, counts of subclasses."""
    summary = {}
    for _ in stats:
        parts = _.split('.')
        klass_name = parts[0]
        if klass_name in ['Meta', 'Extension']:
            continue
        nested_class_count = len(parts[1:])
        total_property_counts = stats[_]
        summary[klass_name] = {'nested_class_count': nested_class_count, 'total_property_counts': total_property_counts}

    stats_table = '\n  '.join(sorted([f'{k}: {v}' for k, v in summary.items()]))
    return '  ' + stats_table


def _update_stats(stats: dict, klass, embedded_class):
    """Update stats for class and embedded class."""
    if embedded_class.__name__ in ['Meta', 'Extension']:
        return
    stat_property = f"{klass.__name__}.{embedded_class.__name__}"
    existing_stat_property = next(iter([k for k in stats.keys() if klass.__name__ in k]), None)
    if existing_stat_property:
        if embedded_class.__name__ not in existing_stat_property:
            property_count = stats[existing_stat_property]
            del stats[existing_stat_property]
            existing_stat_property += f".{embedded_class.__name__}"
            stats[existing_stat_property] = property_count + len(list(embedded_class.element_properties()))
    else:
        stats[stat_property] += 1


def extract_schemas(classes: List[type], base_uri: str) -> dict:
    """Get json schema for all classes."""
    schemas = {}
    for klass in classes:
        schema = klass.schema()

        assert 'title' in schema, schema
        schema['$id'] = schema['title']

        schema = _decorate_schema_bindings(schema)
        schema = _decorate_schema_backrefs(schema)

        schema['description'] = schema.get('description', '').replace('\n\n', '\n')
        schema['description'] = schema['description'].replace("""Disclaimer: Any field name ends with ``__ext`` doesn't part of\nResource StructureDefinition, instead used to enable Extensibility feature\nfor FHIR Primitive Data Types.\n""", '')
        schema['description'] = schema['description'].replace('\n', ' ')

        schema['description'] += f" [See https://hl7.org/fhir/R5/{schema['title']}.html]"

        # rename python style name back to resourceType
        if 'resource_type' in schema['properties']:
            schema['properties']['resourceType'] = schema['properties']['resource_type']
            schema['properties']['resourceType']['description'] = 'One of the resource types defined as part of FHIR'
            del schema['properties']['resource_type']
        # rename python reserved names
        if 'for' in schema['properties']:
            schema['properties']['for_fhir'] = copy.deepcopy(schema['properties']['for'])
            schema['properties']['for_fhir']['description'] = "[Reserved word `for` renamed to `for_fhir`] " + schema['properties']['for_fhir']['description']
            del schema['properties']['for']
        #
        # style as an anonymous schema
        #

        # rename type Resource to $ref
        for p_ in schema['properties'].values():
            if 'type' not in p_:
                continue
            if p_['type'][0].isupper():
                p_['$ref'] = f"{p_['type']}.yaml"
                del p_['type']
        for p_ in schema['properties'].values():
            if 'items' not in p_:
                continue
            if 'type' not in p_['items']:
                continue
            if p_['items']['type'][0].isupper():
                p_['items']['$ref'] = f"{p_['items']['type']}.yaml"
                del p_['items']['type']

        schemas[klass.__name__] = schema

    return schemas


def _decorate_schema_bindings(schema):
    """Supplement schema with bindings."""
    elements = {}
    klass = schema['title']
    with ELEMENT_DB:
        cursor = ELEMENT_DB.cursor()
        cursor.execute('select id, entity from element_bindings where id like ?;', (f"{klass}.%", ))
        for row in cursor.fetchall():
            elements[row[0]] = orjson.loads(row[1])
    if len(elements) == 0:
        return schema
    for property_name, property_ in schema['properties'].items():
        if f"{klass}.{property_name}" not in elements:
            continue
        element = elements[f"{klass}.{property_name}"]
        if 'binding' in element:
            binding = element['binding']
            property_['binding_strength'] = binding.get('strength', None)
            property_['binding_description'] = binding.get('description', None)
            if 'valueSet' in binding:
                uri = binding['valueSet']
                version = None
                if '|' in uri:
                    uri, version = binding['valueSet'].split('|')
                property_['binding_uri'] = uri
                property_['binding_version'] = version
    return schema


def _decorate_schema_backrefs(schema) -> dict:
    """Ensure references have a backref, if there are multiple edges between vertices, distinguish the backref name."""
    references = _scalar_of_type(schema, 'Reference')
    references.update(_scalar_of_type(schema, 'CodeableReference'))
    references.update(_list_of_type(schema, 'Reference'))
    references.update(_list_of_type(schema, 'CodeableReference'))
    targets_to_itemize = _targets_to_itemize(references)
    schema = _decorate_scalar_reference_backref(schema, targets_to_itemize)
    schema = _decorate_list_reference_backref(schema, targets_to_itemize)
    return schema


def _decorate_list_reference_backref(schema: dict, targets_to_itemize: list) -> dict:
    """Set backref in scalar Reference."""
    references = _list_of_type(schema, 'Reference')
    references.update(_list_of_type(schema, 'CodeableReference'))
    if len(references) == 0:
        return schema
    base_backref_name = inflection.underscore(schema['$id'])
    for k, v in references.items():
        backref = f"{k}_{base_backref_name}"
        v['backref'] = backref
    return schema


def _decorate_scalar_reference_backref(schema: dict, targets_to_itemize: list) -> dict:
    """Set backref in scalar Reference."""
    references = _scalar_of_type(schema, 'Reference')
    references.update(_scalar_of_type(schema, 'CodeableReference'))
    if len(references) == 0:
        return schema

    base_backref_name = inflection.underscore(schema['$id'])

    for k, v in references.items():
        enum_reference_types = set(v.get('enum_reference_types', []))
        backref = base_backref_name
        intersection = targets_to_itemize.intersection(enum_reference_types)
        if len(intersection) > 0:
            backref = f"{k}_{base_backref_name}"
            # logger.debug(f"backref: targets:{intersection} {backref}")
        v['backref'] = backref
    return schema


def _targets_to_itemize(references: dict) -> set:
    """Destination vertices with more than one edge."""
    target_counts = defaultdict(int)
    nestedlist = [v.get('enum_reference_types', []) for k, v in references.items()]
    targets = [element for sublist in nestedlist for element in sublist]
    for target in targets:
        target_counts[target] += 1
    targets_to_itemize = set([k for k, v in target_counts.items() if v > 1])
    return targets_to_itemize


def _is_list_of_type(property_: dict, type_: str) -> bool:
    """True if property is a list of type_."""
    if not _is_list_of_any(property_):
        return False
    items = property_.get('items', {})
    return items.get('type', None) == type_ or items.get('$ref', None) == f"{type_}.yaml"


def _is_list_of_any(property_: dict) -> bool:
    """True if property is a list."""
    if not isinstance(property_, dict):
        return False
    if property_.get('type', None) != 'array':
        return False
    items = property_.get('items', None)
    if items is None:
        return False
    items = property_.get('items', {})
    return items.get('type', 'xxx')[0].isupper() or items.get('$ref', 'xxx')[0].isupper()


def _scalar_of_type(schema: dict, type_: str) -> dict:
    """type_ in a scalar field"""
    if 'properties' not in schema:
        return {}
    references = {
        k: v for k, v in schema['properties'].items() if
        isinstance(v, dict) and (v.get('type', None) == type_ or v.get('$ref', None) == f"{type_}.yaml")
    }
    return references


def _list_of_type(schema: dict, type_: str) -> dict:
    """type_ in an array."""
    if 'properties' not in schema:
        return {}
    references = {k: v
                  for k, v in schema['properties'].items() if _is_list_of_type(v, type_)
                  }
    return references


def _list_of_any_type(schema: dict) -> dict:
    """type_ in an array."""
    if 'properties' not in schema:
        return {}
    references = {k: v
                  for k, v in schema['properties'].items() if _is_list_of_any(v)
                  }
    return references


def _scalar_of_any_type(schema: dict) -> dict:
    """Embedded refs"""
    if 'properties' not in schema:
        return {}
    references = {
        k: v for k, v in schema['properties'].items() if
        isinstance(v, dict) and ('.yaml' in v.get('type', '') or '.yaml' in v.get('$ref', ''))
    }
    return references



from jsonschema.exceptions import ValidationError
from jsonschema.validators import Draft202012Validator


def is_edge(schema) -> bool:
    """ Has an edge vocabulary?"""
    vocabulary_fields = [
        'source_property_name',
        'source_multiplicity',
        'destination_property_name',
        'destination_multiplicity',
        'label',
        'is_primary',
        'source_type',
        'destination_type',
    ]
    return all([f in schema for f in vocabulary_fields])


def edge_validator(validator: Draft202012Validator, edge):
    """Extra layer of validation, check both schema validation and vocabulary matching."""
    validator.validate(edge)
    sub_schema = validator.schema['$defs'][edge['resourceType']]
    if is_edge(sub_schema):
        expected_source_type = sub_schema['source_type']
        expected_destination_type = sub_schema['destination_type']
        source = edge['source']
        destination = edge['target']
        if 'reference' in source:
            actual_source_type = source['reference'].split('/')[0]
        else:
            assert 'type' in source, ('Edge.source needs reference or source', edge)
            actual_source_type = source['type']
        if 'reference' in destination:
            actual_destination_type = destination['reference'].split('/')[0]
        else:
            assert 'type' in destination, ('Edge.destination needs reference or source', edge)
            actual_destination_type = destination['type']
        if not actual_source_type == expected_source_type:
            raise ValidationError(f"{actual_source_type} does not match expected {expected_source_type}")
        if not actual_destination_type == expected_destination_type:
            raise ValidationError(f"{actual_destination_type} does not match expected {expected_destination_type}")



import gzip
import io
import logging

import orjson

from iceberg_tools.util import parse_obj

logger = logging.getLogger(__name__)


def _is_bundle(resource):
    """Return True if resource is a bundle."""
    return resource['resourceType'] == 'Bundle'


def _is_fhir_resource(resource):
    """Return True if resource is a FHIR resource."""
    return 'resourceType' in resource


def migrate_resource(resource: dict) -> dict:
    """Apply migrations.

    We solely focus on migrating the necessary migrations for the encountered data, without undertaking a comprehensive migration process.

    :param resource: FHIR resource expecting  https://hl7.org/fhir/r4b/<lower-case-resource-name>
    :return: migrated FHIR resource 5.0 see https://build.fhir.org/<lower-case-resource-name>

    """
    assert 'resourceType' in resource, ('missing resourceType', orjson.dumps(resource).decode())

    resource_type = resource['resourceType']

    migrate_encounter(resource, resource_type)
    migrate_document_reference(resource, resource_type)
    migrate_observation(resource, resource_type)
    migrate_medication_administration(resource, resource_type)
    migrate_research_subject(resource, resource_type)
    migrate_research_study(resource, resource_type)
    migrate_practitioner_role(resource, resource_type)
    migrate_specimen(resource, resource_type)
    migrate_organization(resource, resource_type)
    migrate_location(resource, resource_type)

    return resource


def migrate_location(resource, resource_type):
    """Migrate Location resource to 5.0"""
    if resource_type == "Location":
        contact_0 = {}
        if 'address' in resource:
            contact_0['address'] = resource['address']
            del resource['address']
        if 'telecom' in resource:
            contact_0['telecom'] = resource['telecom']
            del resource['telecom']
        if len(contact_0.keys()) > 0:
            resource['contact'] = [contact_0]


def migrate_organization(resource, resource_type):
    """Migrate Organization resource to 5.0"""
    if resource_type == "Organization":
        contact_0 = {}
        if 'address' in resource:
            contact_0['address'] = resource['address'][0]
            del resource['address']
        if 'telecom' in resource:
            contact_0['telecom'] = resource['telecom']
            del resource['telecom']
        if len(contact_0.keys()) > 0:
            resource['contact'] = [contact_0]


def migrate_specimen(resource, resource_type):
    """Migrate Specimen resource to 5.0"""
    if resource_type == "Specimen":
        if 'collection' in resource:
            if 'bodySite' in resource['collection']:
                resource['collection']['bodySite'] = {
                    'coding': [resource['collection']['bodySite']]
                }
            del resource['collection']


def migrate_practitioner_role(resource, resource_type):
    """Migrate PractitionerRole resource to 5.0"""
    if resource_type == "PractitionerRole":
        if 'telecom' in resource:
            resource['contact'] = [{'telecom': resource['telecom']}]
            del resource['telecom']


def migrate_encounter(resource, resource_type):
    """Migrate Encounter resource to 5.0"""
    if resource_type == "Encounter":
        resource['class'] = [
            {
                'coding': [resource['class']]
            }
        ]
        for _ in resource['participant']:
            if 'individual' in _:
                _['actor'] = _['individual']
                del _['individual']

        if 'period' in resource:
            resource['actualPeriod'] = resource['period']
            del resource['period']

        if 'reasonCode' in resource:
            resource['reason'] = [{'use': resource['reasonCode']}]
            del resource['reasonCode']
        if 'hospitalization' in resource:
            resource['admission'] = resource['hospitalization']
            del resource['hospitalization']


def migrate_document_reference(resource, resource_type):
    """Migrate DocumentReference resource to 5.0"""
    if resource_type == "DocumentReference":
        for _ in resource['content']:
            if 'format' in _:
                del _['format']
        if 'context' in resource and 'encounter' in resource['context']:
            del resource['context']['period']
            resource['context'] = resource['context']['encounter']
        if 'context' in resource and 'related' in resource['context']:
            resource['subject'] = resource['context']['related'][0]
            del resource['context']


def migrate_observation(resource, resource_type):
    """Migrate Observation resource to 5.0"""
    if resource_type == "Observation":
        _ = resource.get('valueSampledData', None)
        if _:
            if 'period' in _:
                _['intervalUnit'] = '/s'
                _['interval'] = _['period']
                del _['period']
        if 'status' not in resource:
            resource['status'] = 'final'


def migrate_medication_administration(resource, resource_type):
    """Migrate MedicationAdministration resource to 5.0"""
    if resource_type == "MedicationAdministration":
        if 'effectiveDateTime' in resource:
            resource['occurenceDateTime'] = resource['effectiveDateTime']
            del resource['effectiveDateTime']

        if 'medicationCodeableConcept' in resource:
            resource['medication'] = {
                'concept': resource['medicationCodeableConcept']
            }
            del resource['medicationCodeableConcept']

        if 'context' in resource:
            resource['encounter'] = resource['context']
            del resource['context']

        if 'reasonReference' in resource:
            resource['reason'] = [{'reference': _} for _ in resource['reasonReference']]
            del resource['reasonReference']


def migrate_research_subject(resource, resource_type):
    """Migrate ResearchSubject resource to 5.0"""
    if resource_type == "ResearchSubject":
        if 'individual' in resource:
            resource['subject'] = resource['individual']
            del resource['individual']


def migrate_research_study(resource, resource_type):
    """Migrate ResearchStudy resource to 5.0"""
    if resource_type == "ResearchStudy":

        if 'category' in resource:
            resource['classifier'] = resource['category']
            del resource['category']

        if 'principalInvestigator' in resource:

            if 'associatedParty' not in resource:
                resource['associatedParty'] = []
            resource['associatedParty'].append(
                {
                    'role': {
                        'coding': [
                            {
                                'system': 'http://hl7.org/fhir/ValueSet/research-study-party-role',
                                'code': 'primary-investigator',
                                'display': 'primary-investigator'
                            }
                        ],
                        'text': 'primary-investigator'
                    },
                    'party': resource['principalInvestigator']
                }
            )
            del resource['principalInvestigator']

        if 'status' not in resource:
            resource['status'] = 'active'

        if 'focus' in resource:
            resource['condition'] = resource['focus']
            del resource['focus']

        if 'enrollment' in resource:
            resource['recruitment'] = resource['enrollment']
            del resource['enrollment']

        # example:
        # "sponsor": {
        #   "reference": "Organization/NLM",
        #   "type": "Organization",
        #   "display": "National Library of Medicine"
        # },
        if 'sponsor' in resource:
            if 'associatedParty' not in resource:
                resource['associatedParty'] = []
            resource['associatedParty'].append(
                {
                    'role': {
                        "coding": [
                            {
                                "system": "http://hl7.org/fhir/ValueSet/research-study-party-role",
                                "code": "sponsor",
                                "display": "sponsor"
                            }
                        ],
                        "text": "sponsor"
                    },
                    'party': {'reference': resource['sponsor']['reference']},
                    'name': 'sponsor'
                }
            )
            del resource['sponsor']

        # example:
        # "recruitment": [
        #   {
        #     "reference": "Group/phs002409.v1.p1-all-subjects"
        #   }
        # ],
        if 'recruitment' in resource:
            resource['recruitment'] = {
                'actualGroup': resource['recruitment'][0]
            }
            del resource['recruitment']


def _is_resource(resource):
    """Return True if resource is a FHIR resource."""
    return 'resourceType' in resource


def migrate_bundles(output_path, path, validate, pattern='**/*.json'):
    """Migrate bundles."""
    for input_file in path.glob(pattern):   # TODO: see util directory_reader
        with open(input_file, "rb") as fp:
            resource = orjson.loads(fp.read())
            if not _is_resource(resource):
                logger.warning(f"Not a FHIR resource {input_file}")
                continue
            if not _is_bundle(resource):
                logger.warning(f"Not a bundle {input_file}")
                continue
            bundle_ = resource
            if 'entry' not in bundle_:
                logger.warning(f"No 'entry' in bundle {input_file} ")
                continue
        for entry in bundle_['entry']:
            resource = entry['resource']
            _ = migrate_resource(resource)
            logging_validator(_, input_file, validate)

        output_file = output_path / input_file.name

        with open(output_file, "wb") as fp:
            fp.write(orjson.dumps(bundle_))
        logger.info(f"Migrated bundle {input_file} to {output_file} entry_count:{len(bundle_['entry'])}")


def migrate_resources(output_path, path, validate, pattern='**/*.json'):
    """Migrate single resource per file."""
    for input_file in path.glob(pattern):  # TODO: see util directory_reader
        with open(input_file, "rb") as fp:
            resource = orjson.loads(fp.read())
            if not _is_resource(resource):
                logger.warning(f"Not a FHIR resource {input_file}")
                continue
            if _is_bundle(resource):
                continue
            _ = migrate_resource(resource)
            logging_validator(_, input_file, validate)

        output_file = output_path / input_file.name
        with open(output_file, "wb") as fp:
            fp.write(orjson.dumps(_))
        logger.info(f'Migrated resource {input_file} to {output_file}')


def logging_validator(_, input_file, validate):
    """Log validation errors."""
    if validate:
        parse_result = parse_obj(_)
        if parse_result.exception is not None:
            if 'resourceType' not in str(parse_result.exception):
                logger.warning(f"{input_file} has exception {parse_result.exception}")


def migrate_ndjson(output_path, path, validate, pattern='**/*.json'):
    """Migrate ndjson files, not expecting bundles."""
    for input_file in path.glob(pattern):  # TODO: see util directory_reader
        out_fp = None
        try:
            with open(input_file, "r") as fp:
                output_file = output_path / input_file.name
                for line in fp.readlines():
                    resource = orjson.loads(line)
                    _ = migrate_resource(resource)
                    if out_fp is None:
                        out_fp = open(output_file, "wb")
                    logging_validator(_, input_file, validate)
                    out_fp.write(orjson.dumps(_, option=orjson.OPT_APPEND_NEWLINE))
            logger.info(f'Migrated {input_file} to {output_file}')
        except orjson.JSONDecodeError as e:
            if 'unexpected end of data' not in str(e):
                logger.warning(f"{input_file} has exception {e}")
        finally:
            if out_fp is not None:
                out_fp.close()


def migrate_json_gz(output_path, path, validate, pattern='**/*.json'):
    """Migrate *json.gz files."""
    if 'gz' not in pattern:
        pattern += ".gz"
    for input_file in path.glob(pattern):  # TODO: see util directory_reader

        with io.TextIOWrapper(io.BufferedReader(gzip.GzipFile(input_file))) as fp:
            output_file = output_path / input_file.name
            with gzip.open(output_file, 'wb') as out_fp:
                for line in fp.readlines():
                    resource = orjson.loads(line)
                    try:
                        _ = migrate_resource(resource)
                        logging_validator(_, input_file, validate)

                        out_fp.write(orjson.dumps(_, option=orjson.OPT_APPEND_NEWLINE))
                    except Exception as e:
                        print('\t', str(e))
                        break
        logger.info(f'Migrated {input_file} to {output_file}')


def migrate_directory(path, output_path, validate, pattern):
    """Migrate all files in a directory."""
    # single resources per file
    migrate_bundles(output_path, path, validate, pattern)
    migrate_resources(output_path, path, validate, pattern)

    # multiple resources per file
    migrate_ndjson(output_path, path, validate, pattern)
    migrate_json_gz(output_path, path, validate, pattern)



import pathlib

import orjson

from iceberg_tools.data.simplifier import simplify_directory, validate_simplified_value


def test_simplify_study():
    """Ensure we can validate a synthetic study"""

    simplify_directory('tests/fixtures/simplify/study/', '**/*.*', 'tmp/study/extractions',
                       'iceberg/schemas/simplified/simplified-fhir.json', 'FHIR', 'config.yaml')

    directory_path = pathlib.Path('tmp/study/extractions')
    input_files = [_ for _ in directory_path.glob("*.ndjson")]
    for file_name in input_files:
        with open(file_name) as fp:
            for line in fp.readlines():
                simplified = orjson.loads(line)
                all_ok = all([validate_simplified_value(_) for _ in simplified.values()])
                assert all_ok, (file_name, line)
                if simplified['resourceType'] == 'DocumentReference':
                    # 'content_md5'
                    expected = set(
                        ['content_attachment_contentType', 'content_attachment_url', 'content_attachment_size',
                         'content_attachment_extension_md5']
                    )
                    actual = set([_ for _ in simplified if _.startswith('content')])
                    assert expected == actual, (file_name, line)


def test_simplify_foo():
    """Ensure we can validate a synthetic study"""

    simplify_directory('tests/fixtures/simplify/foo/', '**/*.*', 'tmp/foo/extractions',
                       'iceberg/schemas/simplified/simplified-fhir.json', 'PFB', 'config.yaml')

    directory_path = pathlib.Path('tmp/foo/extractions')
    input_files = [_ for _ in directory_path.glob("*.ndjson")]
    for file_name in input_files:
        with open(file_name) as fp:
            for line in fp.readlines():
                pfb_ready = orjson.loads(line)
                simplified = pfb_ready['object']
                all_ok = all([validate_simplified_value(_) for _ in simplified.values()])
                print((file_name, line))
                assert all_ok, (file_name, line)
                if simplified['resourceType'] == 'DocumentReference':
                    # 'content_md5'
                    expected = set(
                        ['content_attachment_contentType', 'content_attachment_url', 'content_attachment_size',
                         'content_attachment_extension_md5']
                    )
                    actual = set([_ for _ in simplified if _.startswith('content')])
                    assert expected == actual, (file_name, line)
                    assert len(pfb_ready['relations']) > 0, "DocumentReference should have relationships"



import json
import os
import pathlib

import pytest
from jsonschema.validators import Draft202012Validator

from tests.integration.graph import is_edge


@pytest.fixture()
def bmeg_dir():
    """cd to graph dir"""
    # setup
    cwd = os.getcwd()
    os.chdir('iceberg/schemas/graph')
    # run test
    yield 'iceberg/schemas/graph'
    # teardown
    os.chdir(cwd)


@pytest.fixture()
def fhir_schema(bmeg_dir):
    fhir_schema = json.load(open("graph-fhir.json"))
    Draft202012Validator.check_schema(fhir_schema)
    yield fhir_schema


@pytest.fixture()
def fhir_validator(fhir_schema):
    yield Draft202012Validator(fhir_schema)


@pytest.fixture()
def vertex_schemas(fhir_schema):
    yield [v for v in fhir_schema['$defs'].values() if not is_edge(v)]


@pytest.fixture()
def edge_schemas(fhir_schema):
    yield [v for v in fhir_schema['$defs'].values() if is_edge(v)]


@pytest.fixture()
def lath_path():
    return pathlib.Path('~/go/bin/lathe').expanduser()



import pathlib
from typing import Iterator, List
from zipfile import ZipFile

import orjson
import inflection
import sqlite3
import urllib.request
import logging

logger = logging.getLogger(__name__)


def _class_property(element_id):
    """Ensure type, property_name."""
    parts = element_id.split('.')
    if len(parts) == 2:
        return f"{parts[0]}.{parts[1]}"  # {'type': ''.join(parts[0]), 'property_name': parts[1]}
    else:
        klass = [inflection.camelize(_) for _ in parts[:-1]]
        return f"{''.join(klass)}.{parts[-1]}"  # {'type': ''.join(klass), 'property_name': parts[-1]}


def _profile_in_bundle(bundle) -> Iterator[dict]:
    """Generate profiles."""
    for e in bundle.get('entry', []):
        _ = e.get('resource', None)
        if _:
            yield _


def _profile_in_files(file_names: List[str]) -> dict:
    """Yield the elements with bindings."""
    zip_file = next(iter([_ for _ in file_names if _.endswith('zip')]), None)
    if zip_file:
        with ZipFile(zip_file) as myzip:
            for file_name in [_ for _ in file_names if not _.endswith('zip')]:
                with myzip.open(file_name) as fp:
                    for _ in _profile_in_bundle(orjson.loads(fp.read())):
                        yield _
    else:
        for file_name in file_names:
            with open(file_name) as fp:
                for _ in _profile_in_bundle(orjson.loads(fp.read())):
                    yield _


def _element_in_profile(profile: dict) -> dict:
    """Yield the elements with bindings."""
    snapshot = profile.get('snapshot', None)
    if snapshot:
        for element in snapshot.get('element', []):
            yield element


def _elements_in_files(file_names):
    """Yield the elements with bindings."""
    for profile in _profile_in_files(file_names):
        for element in _element_in_profile(profile):
            yield element


def _elements_with_bindings(file_names: List[str]) -> Iterator[dict]:
    """Elements with bindings, keyed by Class.property"""
    for element in _elements_in_files(file_names):
        if 'binding' not in element:
            continue
        if element['binding'].get('valueSet', None):
            yield {'key': _class_property(element['id']), 'element': element}


def create_fhir_definitions_lookup(iceberg_path: str = "~/.iceberg"):
    """Generate the element_bindings table."""

    path = pathlib.Path(iceberg_path).expanduser()
    path.mkdir(parents=True, exist_ok=True)

    zip_file_path = path / "5.0.0-definitions.json.zip"

    urllib.request.urlretrieve(
        "https://github.com/nazrulworld/hl7-archives/raw/0.4.0/FHIR/R5/5.0.0-definitions.json.zip",
        zip_file_path)

    generator = _elements_with_bindings(
        [str(zip_file_path), 'profiles-resources.json', 'profiles-others.json', 'profiles-types.json'])

    path = path / "element_bindings.sqlite"
    connection = sqlite3.connect(path)

    with connection:
        connection.execute(f'DROP table IF EXISTS element_bindings')
        connection.execute(f'CREATE TABLE if not exists element_bindings (id PRIMARY KEY, entity Text)')
        with connection:
            connection.executemany(f'INSERT OR REPLACE into element_bindings values (?, ?)',
                                   [(_['key'], orjson.dumps(_['element']).decode(),) for _ in generator])
        with connection:
            cursor = connection.cursor()
            cursor.execute('select count(*) from element_bindings;')
            logger.info(f'element_bindings inserted in {path} {cursor.fetchone()[0]}')


if __name__ == "__main__":
    create_fhir_definitions_lookup()



def test_schema_load(fhir_schema, fhir_validator, vertex_schemas, edge_schemas):
    """Ensure whole schema can load"""
    assert fhir_schema
    assert fhir_validator
    assert len(fhir_schema['$defs']) > 90, f"There are {len(fhir_schema['$defs'])} types."
    assert len(vertex_schemas) > 90
    assert len(edge_schemas) == 0



# disability_adjusted_life_years


def test_decimal(distribution_schema: dict):
    """Only json schema types should be used."""
    for entity_name, schema in distribution_schema.items():
        if 'properties' not in schema:
            continue
        if 'identifier' not in schema['properties']:
            continue
        for k, v in schema['properties'].items():
            if not isinstance(v, dict):
                continue
            assert v.get('type', None) != 'decimal', (schema['id'], k)


def test_code(distribution_schema: dict):
    """Only json schema types should be used."""
    for entity_name, schema in distribution_schema.items():
        if 'properties' not in schema:
            continue
        if 'identifier' not in schema['properties']:
            continue
        for k, v in schema['properties'].items():
            if not isinstance(v, dict):
                continue
            assert v.get('type', None) != 'code', (schema['id'], k)



import pytest
from pydantic import ValidationError

from iceberg_tools import monkey_patch_url_validate

monkey_patch_url_validate()

from fhir.resources.attachment import Attachment  # noqa: E402


@pytest.mark.parametrize(
    'value',
    ['file:///foo/bar', 'file://localhost/foo/bar', 'file:////localhost/foo/bar', 'filex://localhost/foo/bar'],
)
def test_valid_file_url(value):
    """Should not raise an exception."""
    a = Attachment.validate({"url": value})
    assert a.url == value, f"{a.url} should be {value}"


@pytest.mark.parametrize(
    'value',
    ['file:foo/bar', 'file:'],
)
def test_invalid_file_url(value):
    """Should raise an exception."""
    with pytest.raises(ValidationError):
        Attachment.validate({"url": value})



import decimal
import importlib
import logging
import pathlib
import threading
import uuid
from typing import Dict, List

import click
import inflection
import orjson
import requests
import yaml

from fhir.resources import FHIRAbstractModel  # noqa
from fhir.resources.attachment import Attachment
from fhir.resources.codeableconcept import CodeableConcept
from fhir.resources.codeablereference import CodeableReference
from fhir.resources.coding import Coding
from fhir.resources.core.utils import is_primitive_type
from fhir.resources.documentreference import DocumentReferenceContent, DocumentReference
from fhir.resources.extension import Extension
from fhir.resources.identifier import Identifier
from fhir.resources.observation import Observation
from fhir.resources.reference import Reference
from fhir.resources.task import Task
from yaml import SafeLoader

from iceberg_tools.util import EmitterContextManager, directory_reader
from iceberg_tools.data.simplifier.oid_lookup import get_oid

# Latest FHIR version by default
FHIR_CLASSES = importlib.import_module('fhir.resources')

LOGGED_ALREADY = []
logger = logging.getLogger(__name__)

# create a local instance
THREAD_LOCAL = threading.local()

ICEBERG_NAMESPACE = uuid.uuid3(uuid.NAMESPACE_DNS, 'ICEBERG')

class Getter:
    """A getter for a FHIR resource."""

    def __init__(self, path: str):
        """Initialize Getter."""
        self.path_str = 'this.' + path

    def _traverse_path(self, resource: FHIRAbstractModel, path_to_search: str = None):
        """Traverse the path, return the last element ,traverse 1st element in list."""
        target = resource
        path_to_search = path_to_search or str(self.path_str)
        path_to_search = path_to_search.replace('this.', '')
        for path in path_to_search.split('.'):
            _ = f"this.{path}"
            _ = eval(_, {}, {'this': target})
            if _ is None:
                break
            if isinstance(_, list):
                target = _[0]
            else:
                target = _
        return _

    def get(self, resource: FHIRAbstractModel):
        """Get the value from the resource, including extensions."""
        if 'extension' in self.path_str:
            prefix, extension_name = self.path_str.split('.extension.')
            val = self._traverse_path(resource, path_to_search=prefix)
            if val:
                extension_list = getattr(val, 'extension', None)
                for extension in extension_list or []:
                    if extension_name in extension.url:
                        return extension.dict()['__value__']
            return None

        return self._traverse_path(resource)


def _get_nested_object(resource: FHIRAbstractModel, path: str):
    """Pluck a value from a resource, using path."""
    return Getter(path).get(resource)


def _simple_coding_dict(self: Coding, *args, **kwargs):
    """MonkeyPatch replacement for dict(), render Coding."""
    display = self.display
    if not display:
        display = self.code
    return [
        {'__value__': f"{self.system}#{self.code}", "__name__": "coding"},
        {'__value__': display},
    ]


def _simple_codeable_concept_dict(self: CodeableConcept, *args, **kwargs):
    """MonkeyPatch replacement for dict(), render CodeableConcept."""
    codeable_concept = []
    if self.text:
        codeable_concept.append(
            {'__value__': [self.text], "__name__": "text"},
        )
    if self.coding is not None:
        codings = [_simple_coding_dict(_) for _ in self.coding]
        # flatten
        codings = [item for sublist in codings for item in sublist]
        codeable_concept.extend(
            [
                {'__value__': [_['__value__'] for _ in codings if _.get('__name__', None) == 'coding'], "__name__": "coding"},
                {'__value__': [_['__value__'] for _ in codings if '__name__' not in _]},
            ]
        )
    return codeable_concept


def _simple_codeable_reference_dict(self: CodeableReference, *args, **kwargs):
    """MonkeyPatch replacement for dict(), render CodeableReference."""
    codeable_reference = []
    if self.reference:
        _ = self.reference.dict()
        _['__value__'] = [_['__value__']]
        codeable_reference.extend([_])
    if self.concept:
        codeable_reference.extend(self.concept.dict())
    return codeable_reference


def _simple_identifier_dict(self: Identifier, *args, **kwargs):
    """MonkeyPatch replacement for dict(), lookup urn:oid values, render Identifier."""
    name = _ensure_identifier_system(self)
    value = _ensure_identifier_value(self)
    identifier = [
        {'__value__': f"{name}#{value}", '__name__': 'identifier'},
        {'__value__': f"{self.system}#{self.value}", '__name__': 'identifier_coding'},
    ]
    if self.type:
        if self.type.text:
            {'__name__': f"{name}_text", '__value__': f"{self.type.text}"}
        if self.type.coding and len(self.type.coding) > 0:
            identifier.append(
                {'__name__': "identifier_text_coding",
                 '__value__': f"{self.type.coding[0].system}#{self.type.coding[0].code}"}
            )
    return identifier


def _ensure_identifier_value(self):
    """Lookup urn: style values. dependent on oid_lookup."""
    value = self.value
    if value:
        if 'urn:oid' in value:
            value_, content = get_oid(value.split('urn:oid:')[-1])
            if not value_:
                _debug_once(('get_oid failed', value, content))
            else:
                value = value_
    return value


def _ensure_identifier_system(self):
    """Ensure identifier `system` simplified."""
    name = None
    if self.system:
        if 'urn:oid' in self.system:
            name, content = get_oid(self.system.split('urn:oid:')[-1])
            if not name:
                _debug_once(('get_oid failed', self.system, content))
        elif 'urn:ietf:rfc:3986' in self.system:
            name = 'uri'
        else:
            name = self.system.split('/')[-1].replace('-', '_')
    if not name:
        name = ''
    return name


def _simple_observation_dict(self: Observation, *args, **kwargs):
    """MonkeyPatch replacement for dict(), render Observation."""
    observation = _simple_resource_dict(self)
    # TODO extend for specific observations
    if 'category' not in observation or observation['category'] == [None]:
        observation['category'] = ['laboratory']
        observation['category_coding'] = ['http://terminology.hl7.org/CodeSystem/observation-category#laboratory']
    return observation


def _simple_extension_dict(self: Extension, *args, **kwargs):
    """MonkeyPatch replacement for dict(), render Extension."""
    # note `self` is the Extension
    # stop traversal here
    alias_maps = self.get_alias_mapping()
    for prop_name in self.elements_sequence():
        field_key = alias_maps[prop_name]
        is_value = field_key.startswith('value')
        if is_value:
            v = self.__dict__.get(field_key, None)
            if v is not None:
                is_fhir = issubclass(type(v), FHIRAbstractModel)
                _name = self.url
                if _name:
                    _name = _name.split('/')[-1]
                _name = _name.replace('-', '_')
                if is_fhir:
                    v = v.dict()
                _ = {'__value__': v}
                if _name:
                    _['__name__'] = _name
                return _


def _simple_reference_dict(self: Reference, *args, **kwargs):
    """MonkeyPatch replacement for dict(), render Reference."""
    # note `self` is the Reference
    # stop traversal here
    THREAD_LOCAL.references.append(self.reference)
    return {'__value__': self.reference}


def _simple_attachment_dict(self: Attachment, *args, **kwargs):
    """MonkeyPatch replacement for dict(), render Attachment."""
    attachment = _render_extensions(self)
    attachment.update(_render_primitives(self))
    return attachment


def _simple_document_reference_content_dict(self: DocumentReferenceContent, *args, **kwargs):
    """MonkeyPatch replacement for dict(), render DocumentReferenceContent."""
    document_reference_content = _render_extensions(self)
    document_reference_content.update(_render_primitives(self))
    document_reference_content.update(self.attachment.dict())
    return document_reference_content


def _simple_document_reference_dict(self: DocumentReference, *args, **kwargs):
    """MonkeyPatch replacement for dict(), render DocumentReference."""
    # defaults
    document_reference = _simple_resource_dict(self)
    return document_reference


def _simple_task_dict(self: Task, *args, **kwargs):
    """MonkeyPatch replacement for dict(), render Observation."""
    task = _simple_resource_dict(self)
    # go through outputs to collect references
    for _ in self.output:
        _.dict()
    return task


def _render_sub_resources(resource: FHIRAbstractModel) -> dict:
    """Render sub-resources with keys "{sub_resource}_{if list index}_{property_name}: simple_value" """

    # read config
    nested_objects = THREAD_LOCAL.nested_objects
    # render lists to scalar
    simplified = {}

    if resource.resource_type not in nested_objects:
        return simplified

    paths = nested_objects[resource.resource_type]
    for path in paths:
        nested_object_value = _get_nested_object(resource=resource, path=path)
        nested_object_name = '_'.join(path.split('.'))

        # # fyi:resource.schema() caches results
        # parent_property_name = path.split('.')[0]
        # parent_property_schema = resource.schema()['properties'][parent_property_name]

        is_array = nested_object_value is not None and isinstance(nested_object_value, list)

        if is_array and nested_object_value is not None:

            count = 0
            for item in nested_object_value:
                _ = item.dict()

                # only add index in name if > 0
                separator = '_'

                value = _['__value__']
                name = _.get('__name__', None)
                if name:
                    name = separator + name
                else:
                    name = ''

                if f"{nested_object_name}{name}" not in simplified:
                    simplified[f"{nested_object_name}{name}"] = []
                simplified[f"{nested_object_name}{name}"].append(value)

                count += 1

        else:
            # scalar
            if nested_object_value is not None:
                if hasattr(nested_object_value, 'dict'):
                    name_value_list = nested_object_value.dict()
                    if not isinstance(name_value_list, list):
                        name_value_list = [name_value_list]
                    for item in name_value_list:
                        value = item['__value__']
                        name = item.get('__name__', None)
                        if name:
                            name = '_' + name
                        else:
                            name = ''
                        if isinstance(value, list):
                            value = value[0]
                        simplified[f"{nested_object_name}{name}"] = value
                else:
                    if isinstance(nested_object_value, list):
                        nested_object_value = nested_object_value[0]
                    simplified[f"{nested_object_name}"] = nested_object_value

    return simplified


def _simple_resource_dict(self: FHIRAbstractModel, *args, **kwargs):
    """MonkeyPatch replacement for dict(), render generic Resource."""
    # override this with a specialization
    simplified = _render_extensions(self)
    simplified.update(_render_primitives(self))
    simplified.update(_render_codings(self))
    simplified.update(_render_identifiers(self))
    simplified.update(_render_references(self))
    simplified.update(_render_sub_resources(self))
    return simplified


def _render_primitives(self: FHIRAbstractModel, *args, **kwargs) -> Dict:
    """Utility, create a dict of the primitive keys in a resource"""
    alias_maps = self.get_alias_mapping()
    resource_primitives = {}
    for prop_name in self.elements_sequence():
        field_key = alias_maps[prop_name]
        is_primitive = is_primitive_type(self.__fields__[field_key])
        v = self.__dict__.get(field_key, None)
        if v is not None and is_primitive:
            resource_primitives[field_key] = v
    if 'resourceType' not in resource_primitives:
        resource_primitives['resourceType'] = self.resource_type
    return resource_primitives


def _render_references(self: FHIRAbstractModel, *args, **kwargs) -> Dict:
    """Utility, create a dict of the references in a resource."""
    alias_maps = self.get_alias_mapping()
    references = {}
    for prop_name in self.elements_sequence():
        field_key = alias_maps[prop_name]
        v = self.__dict__.get(field_key, None)
        if v is not None and isinstance(v, Reference):
            references[field_key] = v.dict()['__value__']
    return references


def _render_extensions(self: FHIRAbstractModel, *args, **kwargs) -> Dict:
    """Utility, create a dict of the extensions in a resource."""
    extensions = {}
    if hasattr(self, 'extension') and self.extension:
        for extension in self.extension:
            ext_dict = extension.dict()
            if ext_dict:
                _populate_extensions(ext_dict, extensions)
            else:
                parent_name = extension.url.split('/')[-1].replace('-', '_')
                for child_extension in extension.extension:
                    ext_dict = child_extension.dict()
                    if ext_dict:
                        _populate_extensions(ext_dict, extensions, parent_name)

    # cleanup no data
    extensions = {k: v for k, v in extensions.items() if (v is not None and v != [])}
    return extensions


def _populate_extensions(ext_dict, extensions, parent_name=None):
    """Utility, populate extensions dict with values from a single extension."""
    ext_val_as_list = ext_dict['__value__']
    if not isinstance(ext_dict['__value__'], list):
        ext_val_as_list = [ext_dict['__value__']]
    ext_name = ext_dict['__name__']
    if parent_name:
        ext_name = f"{parent_name}_{ext_name}"

    compound_name = None
    for item in ext_val_as_list:
        # more than one name
        compound_name = ext_name

        if not isinstance(item, dict):
            extensions[compound_name] = item
            continue

        if '__name__' in item and item['__name__'] != ext_name:
            compound_name = f"{compound_name}_{item['__name__']}"
        compound_name = compound_name.replace('-', '_')

        compound_name = compound_name.replace('_text', '')

        if compound_name not in extensions:
            extensions[compound_name] = []

        if '__value__' in item:
            extensions[compound_name].append(item['__value__'])
        else:
            for k, v in item.items():
                extensions[f"{compound_name}_{k}"] = v

    # if only one value, make it scalar
    if isinstance(extensions[compound_name], list) and len(extensions[compound_name]) == 1:
        extensions[compound_name] = extensions[compound_name][0]


def _render_codings(self: FHIRAbstractModel, *args, **kwargs) -> Dict:
    """Utility, create a dict of the codings in a resource."""
    codings = {}
    alias_maps = self.get_alias_mapping()
    for prop_name in self.elements_sequence():
        field_key = alias_maps[prop_name]
        v = self.__dict__.get(field_key, None)
        is_scalar = not isinstance(v, list)

        if v is not None:
            v_as_list = v
            if not isinstance(v_as_list, list):
                v_as_list = [v_as_list]
            if len(v_as_list) == 0:
                continue
            if type(v_as_list[0]) not in [CodeableConcept, Coding, CodeableReference]:
                continue
            for v in v_as_list:
                codings_as_list = v.dict()
                if not isinstance(codings_as_list, list):
                    codings_as_list = [codings_as_list]
                for _ in codings_as_list:
                    compound_name = field_key
                    if '__name__' in _:
                        compound_name = f"{compound_name}_{_['__name__']}"
                    compound_name = compound_name.replace('-', '_')
                    if compound_name not in codings:
                        codings[compound_name] = []
                    codings[compound_name].extend(_['__value__'])
                    if is_scalar:
                        codings[compound_name] = codings[compound_name][0]
    return codings


def _render_identifiers(self: FHIRAbstractModel, *args, **kwargs) -> Dict:
    """Utility, create a dict of the identifiers in a resource."""

    identifiers = {}

    if hasattr(self, 'identifier') and self.identifier:

        # treat as a list, Per FHIR most are, some are not
        identifier_list = self.identifier
        if not isinstance(self.identifier, list):
            identifier_list = [self.identifier]

        # render as {k:[]}
        rendered_identifiers = []
        for _ in identifier_list:
            rendered_identifiers.extend(_simple_identifier_dict(_))
        keys = [_['__name__'] for _ in rendered_identifiers]
        for k in keys:
            identifiers[k] = [_['__value__']for _ in rendered_identifiers if _['__name__'] == k]

    return identifiers


def validate_simplified_value(v) -> bool:
    """Is a field valid in a simplified model?"""
    if v is None:
        return True
    if isinstance(v, list) and len(v) == 0:
        return True
    if isinstance(v, list) and issubclass(type(v[0]), FHIRAbstractModel):
        return False
    if isinstance(v, list) and isinstance(v[0], dict):
        return False
    if issubclass(type(v), FHIRAbstractModel):
        return False
    if isinstance(v, dict):
        return False
    # if isinstance(v, bytes):
    #     return False
    return True


def _gen3_scaffolding_document_reference(simplified: dict, resource: FHIRAbstractModel) -> dict:
    """Add Gen3 boiler plate"""

    if simplified['resourceType'] != 'DocumentReference':
        return simplified

    if 'content_attachment_url' not in simplified:
        logger.debug(f"Missing content_attachment_url in simplified {simplified['resourceType']} {simplified['id']}")
        suffix = None
        file_name = None
    else:
        content_url = pathlib.Path(simplified['content_attachment_url'])
        suffix = content_url.suffix
        file_name = content_url.name

    if suffix:
        suffix = suffix.replace('.', '')
    simplified['data_type'] = suffix
    if suffix in ['csv']:
        simplified['data_format'] = 'variants'
    if suffix in ['dcm', 'tif']:
        simplified['data_format'] = 'imaging'
    if suffix in ['txt']:
        simplified['data_format'] = 'note'
    if suffix in ['out', 'log']:
        simplified['data_format'] = 'log'
    if suffix in ['m']:
        simplified['data_format'] = 'matlab'
    if suffix in ['sh']:
        simplified['data_format'] = 'script'

    simplified['file_name'] = file_name
    if 'content_attachment_extension_md5' in simplified:
        simplified['md5sum'] = simplified['content_attachment_extension_md5']
    if 'content_size' in simplified:
        simplified['file_size'] = simplified['content_attachment_size']
    simplified['object_id'] = simplified['id']

    return simplified


def _gen3_references(simplified: dict, resource: FHIRAbstractModel) -> dict:

    pass


def _ensure_dialect(simplified: dict, resource: FHIRAbstractModel, dialect: str) -> dict:
    """Enforce dialect rules, update simplified dict.
    In practice, it toggles between PFB and plain FHIR.
    PFB dialect injects Gen3 properties and harvests edges.
    """
    if dialect != 'PFB':
        return simplified
    simplified = _gen3_scaffolding_document_reference(simplified, resource)
    return simplified


def _render_dialect(simplified: dict, references: List[str], dialect: str, schemas: dict, limit_links: dict = {}) -> dict:
    """Render as PFB record ready for import
    """
    if dialect != 'PFB':
        return simplified

    labels = [_['title'] for _ in schemas.values() if 'title' in _]

    permitted_destinations = None
    if simplified['resourceType'] in limit_links:
        permitted_destinations = limit_links[simplified['resourceType']]

    gen3_links = []
    for _ in references:
        if len(_.split('/')) != 2:
            logger.warning(f"Unexpected reference format {_} in {simplified['resourceType']} {simplified['id']}")
            continue
        label, id_ = _.split('/')
        if label not in labels:
            continue
        if permitted_destinations is not None and label not in permitted_destinations:
            continue
        gen3_links.append({"dst_id": id_, "dst_name": inflection.underscore(label)})

    return {'id': simplified['id'], 'name': inflection.underscore(simplified['resourceType']), 'relations': gen3_links, 'object': simplified}


def new_id(transform_ids, id_):
    """Create a new UUID based on the original id and the seed."""
    _study_uuid = uuid.uuid5(ICEBERG_NAMESPACE, transform_ids)
    return str(uuid.uuid5(_study_uuid, id_))


def simplify(resource: FHIRAbstractModel, dialect: str, nested_objects: dict = {}, transform_ids=None) -> (Dict, List[str]):
    """Create a PFB friendly resource. Returns simplified resource and associated references.
    Most of the heavy lifting done in dict() overrides of FHIRAbstractModel.dict()
    """
    # filled in by _simple_reference_dict
    THREAD_LOCAL.references = []
    # read by _render_sub_resources
    THREAD_LOCAL.nested_objects = nested_objects
    simplified = resource.dict()
    # cleanup goes here...
    simplified = _ensure_dialect(simplified, resource, dialect)
    if transform_ids:
        simplified['id'] = new_id(transform_ids, simplified['id'])
        transformed_references = []
        for reference in THREAD_LOCAL.references:
            type_, id_ = reference.split('/')
            id_ = new_id(transform_ids, id_)
            transformed_references.append(f"{type_}/{id_}")
        THREAD_LOCAL.references = transformed_references
    return simplified, THREAD_LOCAL.references


def check_simplified_schemas(simplified: dict, schemas: dict):
    """Compare simplified fields vs schema."""
    resource_type = f"{inflection.underscore(simplified['resourceType'])}.yaml"
    assert resource_type in schemas, f"{resource_type} not in schemas"
    # TODO - extra fields are in $ref, need to resolve
    # extra_fields = [k for k in simplified if k not in schemas[resource_type]['properties']]
    # assert len(extra_fields) == 0, ('extra_fields_not_in_schema', resource_type, extra_fields)


def _default_json_serializer(obj):
    """JSON Serializer, render decimal and bytes types."""
    if isinstance(obj, decimal.Decimal):
        return float(obj)
    if isinstance(obj, bytes):
        return obj.decode()

    raise TypeError


class SimplifierContextManager:
    """Monkey patch FHIR classes, restore when complete"""
    def __enter__(self):
        # print("Monkey patching FHIR", file=sys.stderr)
        self.orig_resource_dict = FHIRAbstractModel.dict
        self.orig_coding_dict = Coding.dict
        self.orig_identifier_dict = Identifier.dict
        self.orig_reference_dict = Reference.dict
        self.orig_extension_dict = Extension.dict

        self.orig_attachment_dict = Attachment.dict
        self.orig_document_reference_content_dict = DocumentReferenceContent.dict

        self.orig_observation_dict = Observation.dict
        self.orig_codeable_concept_dict = CodeableConcept.dict
        self.orig_codeable_reference_dict = CodeableReference.dict

        self.orig_task_dict = Task.dict

        FHIRAbstractModel.dict = _simple_resource_dict
        Coding.dict = _simple_coding_dict
        Identifier.dict = _simple_identifier_dict
        Reference.dict = _simple_reference_dict
        Extension.dict = _simple_extension_dict

        # Attachment.dict = _simple_attachment_dict
        # DocumentReferenceContent.dict = _simple_document_reference_content_dict

        Observation.dict = _simple_observation_dict
        CodeableConcept.dict = _simple_codeable_concept_dict
        CodeableReference.dict = _simple_codeable_reference_dict

        Task.dict = _simple_task_dict    # TODO - do we still need this?

    def __exit__(self, exc_type, exc_value, exc_tb):
        # print("Restoring original FHIR classes", file=sys.stderr)
        FHIRAbstractModel.dict = self.orig_resource_dict
        Coding.dict = self.orig_coding_dict
        Identifier.dict = self.orig_identifier_dict
        Reference.dict = self.orig_reference_dict
        Extension.dict = self.orig_extension_dict

        Attachment.dict = self.orig_attachment_dict
        DocumentReferenceContent.dict = self.orig_document_reference_content_dict

        Observation.dict = self.orig_observation_dict
        CodeableConcept.dict = self.orig_codeable_concept_dict
        CodeableReference.dict = self.orig_codeable_reference_dict

        Task.dict = self.orig_task_dict


def simplify_directory(input_path, pattern, output_path, schema_path, dialect, config_path, transform_ids=None):
    """Reads directory of FHIR, renders simple, data frame friendly flattened records."""

    input_path = pathlib.Path(input_path)
    assert input_path.is_dir(), f"{input_path} not a directory"
    dialect = dialect.upper()

    if pathlib.Path(schema_path).is_file():
        with open(schema_path, "rb") as fp_:
            schemas = orjson.loads(fp_.read())
    else:
        schemas = requests.get(schema_path).json()
    assert schemas, f"No schema found at {schema_path}"

    with open(config_path) as fp:
        gen3_config = yaml.load(fp, SafeLoader)
    limit_links = gen3_config['limit_links']
    nested_objects = gen3_config['nested_objects']

    with SimplifierContextManager():
        with EmitterContextManager(output_path) as emitter:
            for parse_result in directory_reader(directory_path=input_path, pattern=pattern,
                                                 validate=False, ignore_path=output_path):
                if parse_result.exception is not None:
                    if 'resourceType' not in str(parse_result.exception):
                        logger.error(f"{parse_result.path} has exception {parse_result.exception}")
                    continue
                resource = parse_result.resource

                simplified, references = simplify(resource, dialect, nested_objects, transform_ids)
                try:
                    check_simplified_schemas(simplified, schemas)
                except (TypeError, AssertionError) as e:
                    _debug_once(str(e))

                assert simplified, ("Should have simplified", resource.resource_type, resource.id)
                all_ok = all([validate_simplified_value(_) for _ in simplified.values()])
                _assert_all_ok(all_ok, parse_result, resource, simplified)

                simplified = _render_dialect(simplified, references, dialect, schemas, limit_links)
                fp = emitter.emit(resource.resource_type)
                fp.write(orjson.dumps(simplified, default=_default_json_serializer,
                                      option=orjson.OPT_APPEND_NEWLINE).decode())


def _debug_once(msg):
    if msg not in LOGGED_ALREADY:
        LOGGED_ALREADY.append(msg)
        logger.info(msg)


def _assert_all_ok(all_ok, parse_result, resource, simplified):
    """Utility, log simplified check problems."""
    if not all_ok:
        reference = f"{resource.resource_type}/{resource.id}"
        logger.warning(f"{parse_result.path} {reference} {parse_result.offset}")
        logger.warning(f"\t{simplified}")
        logger.warning({k: validate_simplified_value(v) for k, v in simplified.items()})
    assert all_ok


@click.command('simplify')
@click.argument('path')
@click.argument('output_path')
@click.option('--pattern', required=True, default="**/*.*", show_default=True,
              help='File name pattern')
@click.option('--schema_path', required=True,
              show_default=True,
              help='Path to simplified schema json.  (Accepts url or file path)'
              )
@click.option('--dialect',
              default='PFB',
              type=click.Choice(['FHIR', 'PFB'], case_sensitive=False),
              help='PFB: adds common properties, FHIR: passthrough'
              )
@click.option('--config_path',
              default='config.yaml',
              show_default=True,
              help='Path to config file.')
@click.option('--transform_ids',
              default=None,
              show_default=True,
              help='Transform ids based on this seed')

def cli(path, pattern, output_path, schema_path, dialect, config_path, transform_ids):
    """Renders PFB friendly flattened records.

    PATH: Path containing bundles (*.json) or resources (*.ndjson)
    OUTPUT_PATH: Path where simplified resources will be stored
    """
    simplify_directory(path, pattern, output_path, schema_path, dialect, config_path, transform_ids)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    cli()



import logging
import pathlib
import threading

import click
import yaml
from fhir.resources import FHIRAbstractModel  # noqa
from yaml import SafeLoader

from iceberg_tools.data.report import aggregate_edges
from iceberg_tools.data.migrator import migrate_directory
from iceberg_tools.data.pfb import SimplePFBWriter
from iceberg_tools.data.simplifier import cli as simplifier
from iceberg_tools.schema.simplified_validator import directory_reader as simplified_directory_reader
from iceberg_tools.util import NaturalOrderGroup, directory_reader

LINKS = threading.local()
CLASSES = threading.local()
IDENTIFIER_LIST_SIZE = 8

logger = logging.getLogger(__name__)


@click.group(name="data", cls=NaturalOrderGroup)
def cli():
    """Project data (ResearchStudy, ResearchSubjects, Patient, etc.)."""
    pass


cli.add_command(simplifier)


@cli.command('validate')
@click.argument('path')
@click.option('--pattern', required=True, default="*.*", show_default=True,
              help='File name pattern')
def _validate(path, pattern):
    """Check FHIR data for validity and conventions.

    PATH: Path to FHIR files.
    """
    ok = True
    for result in directory_reader(pathlib.Path(path), pattern):
        if result.exception:
            ok = False
            print('file:', result.path)
            print('\toffset:', result.offset)
            print('\tresource_id:', result.resource_id)
            msg = str(result.exception).replace('\n', '\n\t\t')
            print('\texception:', msg)
    if ok:
        print('OK, all resources pass')


@cli.command('validate-simplified')
@click.argument('path')
@click.option('--schema_path', required=True,
              default='iceberg/schemas/simplified/simplified-fhir.json',
              show_default=True,
              help='Path to simplified schema json, a file path or url'
              )
def _validate_simplified(path, schema_path):

    """Check simplified data for validity and conventions.

    PATH: Path to simplified ndjson files.
    """
    ok = True
    for result in simplified_directory_reader(pathlib.Path(path), schema_path):
        if result.exception:
            ok = False
            print('file:', result.path)
            print('\toffset:', result.offset)
            print('\tresource_id:', result.resource_id)
            msg = str(result.exception).replace('\n', '\n\t\t')
            print('\texception:', msg)
    if ok:
        print('OK, all resources pass')


@cli.command('pfb')
@click.argument('path')
@click.argument('output_path')
@click.option('--schema_path', required=True,
              default='iceberg/schemas/simplified/simplified-fhir.json',
              show_default=True,
              help='Path to simplified schema json, a file path or url'
              )
@click.option('--config_path',
              default='config.yaml',
              show_default=True,
              help='Path to config file.')
def pfb(path, output_path, schema_path, config_path):

    """Write simplified FHIR files to a PFB.

    \b
    PATH: Directory path to simplified FHIR ndjson files.
    OUTPUT_PATH: File path where to write the PFB.
    """
    path = pathlib.Path(path)
    output_path = pathlib.Path(output_path)
    config_path = pathlib.Path(config_path)
    assert config_path.is_file(), f"Path {config_path} is not a file"
    assert path.is_dir(), f"Path {path} is not a directory"

    with open(config_path) as fp:
        gen3_config = yaml.load(fp, SafeLoader)
    dependency_order = gen3_config['dependency_order']

    pfb_writer = SimplePFBWriter(schema_path=schema_path,
                                 output_path=output_path,
                                 dependency_order=dependency_order
                                 )

    # this is a generator, needs to be consumed
    [_ for _ in pfb_writer.transform_directory(path)]

    inspection = pfb_writer.inspect()
    if len(inspection.errors) > 0:
        logger.error(inspection.warnings)
    if len(inspection.warnings) > 0:
        logger.warning(inspection.warnings)
    if len(inspection.info) > 0:
        for _ in inspection.info:
            logger.info(_)


@cli.command('migrate')
@click.argument('path')
@click.argument('output_path')
@click.option('--validate/--no-validate', default=True, is_flag=True, show_default=True,
              help="Validate after migration")
@click.option('--pattern', required=True, default="**/*.*json", show_default=True,
              help='File name pattern')
def migrate(path, output_path, validate, pattern):
    """Migrate from FHIR R4B to R5.0.


    We solely focus on migrating the necessary migrations for the encountered data, without undertaking a comprehensive migration process.
    Please refer to repository for more details: https://github.com/bmeg/iceberg-schema-tools

    PATH: Path containing bundles (*.json) or resources (*.ndjson)
    OUTPUT_PATH: Path where migrated resources will be stored
    """

    path = pathlib.Path(path)
    output_path = pathlib.Path(output_path)
    assert path.is_dir(), f"Path {path} is not a directory"

    if not output_path.exists():
        output_path.mkdir(parents=True, exist_ok=True)

    assert output_path.is_dir(), f"Path {output_path} is not a directory"

    migrate_directory(path, output_path, validate, pattern)


@cli.command()
@click.argument('path')
@click.argument('output_path')
@click.option('--pattern', default='**/*.avro', help='Search pattern', show_default=True)
def report(path: str,  output_path: str,  pattern: str):
    """Aggregate avro pfb files into a cytoscape tsv.

    \b
    PATH: Directory path to search for pfb files
    OUTPUT_PATH: Directory path to write the report TSVs
    """
    assert pathlib.Path(path).is_dir(), f"Path {path} is not a directory"
    assert pathlib.Path(output_path).is_dir(), f"Path {output_path} is not a directory"
    aggregate_edges(path, output_path, pattern)


if __name__ == '__main__':
    cli()




def test_research_study_has_project(distribution_schema: dict):
    """Ensure there is a link between ResearchSubject and Project"""
    assert "research_study.yaml" in distribution_schema
    research_study = distribution_schema["research_study.yaml"]
    assert "_project.yaml" in distribution_schema
    link_labels = [_['label'] for _ in research_study['links']]
    assert 'ResearchStudy_gen3_project_Project_research_study' in link_labels, link_labels
    assert '_project' not in research_study['properties']



from dictionaryutils import DataDictionary


def test_keys(data_dictionary_from_yaml: DataDictionary):
    """Schema has Gen3 friendly identifiers."""
    schema = data_dictionary_from_yaml.schema
    assert all([_.islower() for _ in schema]), "All keys should be lowercase"



import json
import os

import pathlib
from typing import Any

import pytest
import requests
import yaml

import fastjsonschema

from iceberg_tools.graph import VertexLinkWriter
from iceberg_tools.schema.graph import bundle_schemas

SCHEMA_PATHS = """
        tests/fixtures/swapi/character.yaml
        tests/fixtures/swapi/film.yaml
        tests/fixtures/swapi/planet.yaml
        tests/fixtures/swapi/species.yaml
        tests/fixtures/swapi/starship.yaml
        tests/fixtures/swapi/vehicle.yaml
        tests/fixtures/swapi/planets.yaml
        tests/fixtures/swapi/reference.yaml
    """.split()

RESOURCES = {}


def load_yaml(path) -> dict:
    """Load yaml file and return dict.
    :param path: path to yaml file
    """
    with open(path) as fp:
        return yaml.safe_load(fp)


def retriever(uri: str):
    """Retrieve a resource from RESOURCES or the internet."""

    global RESOURCES
    if not RESOURCES:
        for path in SCHEMA_PATHS:
            name = pathlib.Path(path).name
            schema = load_yaml(path)
            RESOURCES[name] = schema

    if uri in RESOURCES:
        return RESOURCES[uri]

    return requests.get(uri).json()


def compile_schema(schema) -> Any:
    """Compile schema dict with $ref resolver"""
    return fastjsonschema.compile(schema, handlers={'': retriever, 'http': retriever, 'https': retriever})


def validate_hyper_schema(schema) -> dict:
    """Load yaml file for schema and ensure it is valid.
    :param schema: schema dict
    Raises exception if schema is not valid. (SchemaError, ValidationError
    """

    # validate it is a valid jsonschema
    return compile_schema(schema)


def type_is(property_, type_) -> bool:
    """Check if property[type] is of type_."""
    if "type" not in property_:
        return False
    if type_ == property_["type"]:
        return True
    if type_ in property_["type"]:
        return True
    return False


def test_validator():
    """Test the validator."""
    for path in SCHEMA_PATHS:
        print("Validating", path)
        schema = load_yaml(path)
        validate_hyper_schema(schema)
        print(f"\tOK {schema['title']} schema is valid")
        break


def test_valid_schemas():
    """Load yaml files for all schema and ensure it is valid."""

    for path in SCHEMA_PATHS:
        print("Validating", path)

        schema = load_yaml(path)
        validate_hyper_schema(schema)
        print(f"\tOK {schema['title']} schema is valid")

        # ensure that link rels are unique
        rel_names = [_['rel'] for _ in schema.get("links", [])]
        assert len(rel_names) == len(set(rel_names)), f"{schema['title']} has duplicate link rels"

        # ensure that every link is also a property
        for link in schema.get("links", []):
            assert link["rel"] in schema["properties"], f"{link['rel']} not in {schema['properties']}"

            print(f"\tOK {schema['title']} link.{link['rel']} has properties entry")

        # ensure that every link multiplicity is appropriate
        for link in schema.get("links", []):
            property_ = schema["properties"][link["rel"]]
            target_hints = link.get("targetHints", {})
            multiplicity = target_hints.get("multiplicity", "has_many")
            if multiplicity == "has_many":
                assert type_is(property_,
                               "array"), f"{link['rel']} has multiplicity has_many and therefore should be an array"
            elif multiplicity == "has_one":
                assert not type_is(property_, "array"), f"{link['rel']} has_one should not be an array"
            else:
                raise ValueError(f"Unknown multiplicity {multiplicity} for {link['rel']}")

            print(f"\tOK {schema['title']} link.{link['rel']} multiplicity to property type")


def test_instances():
    """Ensure sample instances are valid."""

    #     tests/fixtures/swapi/swapi_planets.ndjson association
    instances = """tests/fixtures/swapi/swapi_character.ndjson
    tests/fixtures/swapi/swapi_film.ndjson
    tests/fixtures/swapi/swapi_planet.ndjson
    tests/fixtures/swapi/swapi_species.ndjson
    tests/fixtures/swapi/swapi_starship.ndjson
    tests/fixtures/swapi/swapi_vehicle.ndjson
    tests/fixtures/swapi/swapi_planets.ndjson
    """.split()

    for path in instances:
        print(f"Validating instances in: {path}")
        with open(path) as fp:
            validate = None
            for line in fp:
                instance = json.loads(line)
                if validate is None:
                    _ = load_yaml(f"tests/fixtures/swapi/{instance['label'].lower()}.yaml")
                    validate = compile_schema(_)
                # print(f"\tvalidating:", instance)
                validate(instance)
                print(f"\tOK {instance['gid']} instance is valid")


def test_generate_links_from_instances():
    """Ensure we can generate links from instances + schemas, test links uniqueness."""

    instances = """tests/fixtures/swapi/swapi_character.ndjson
    tests/fixtures/swapi/swapi_film.ndjson
    tests/fixtures/swapi/swapi_planet.ndjson
    tests/fixtures/swapi/swapi_species.ndjson
    tests/fixtures/swapi/swapi_starship.ndjson
    tests/fixtures/swapi/swapi_vehicle.ndjson
    tests/fixtures/swapi/swapi_planets.ndjson
    """.split()

    for path in instances:
        print(f"Generating links for instances in: {path}")
        schema = None
        with open(path) as fp:
            for line in fp:
                instance = json.loads(line)
                schema = load_yaml(f"tests/fixtures/swapi/{instance['label'].lower()}.yaml")
                break

        with VertexLinkWriter(schema) as mgr:
            with open(path) as fp:
                for line in fp:
                    instance = mgr.insert_links(json.loads(line))

                    assert 0 < len(instance['links']), f"{instance['gid']} does not have at least one link"

                    assert all(['href' in _ and 'rel' in _ for _ in
                                instance['links']]), f"{instance['gid']} links do not have href and rel"

                    instance_links = [_['href'] for _ in instance['links']]
                    print(sorted(instance_links))
                    if len(sorted(instance_links)) != len(set(sorted(instance_links))):
                        print("?")
                    assert len(sorted(instance_links)) == len(set(sorted(instance_links))), f"{instance['gid']} has duplicate links {sorted(instance_links)}"
                    print(f"\tOK {instance['gid']} has {len(instance['links'])} links")


@pytest.mark.skip(reason="Use to create bundled json file")
def test_bundled():
    """Ensure we can bundle yaml files into a single json file."""

    bundle_schemas(pathlib.Path("tests/fixtures/swapi"), "http://bmeg-swapi/0.0.1", "graph-swapi.json")
    validate_hyper_schema(json.load(open("tests/fixtures/swapi/graph-swapi.json")))


@pytest.mark.skip(reason="Use to create edge json file")
def test_create_links():
    """Ensure we can create instances with links."""

    instances = """tests/fixtures/swapi/swapi_character.ndjson
    tests/fixtures/swapi/swapi_film.ndjson
    tests/fixtures/swapi/swapi_planet.ndjson
    tests/fixtures/swapi/swapi_species.ndjson
    tests/fixtures/swapi/swapi_starship.ndjson
    tests/fixtures/swapi/swapi_vehicle.ndjson
    tests/fixtures/swapi/swapi_planets.ndjson
    """.split()

    for path in instances:
        print(f"Generating links for instances in: {path}")
        schema = None
        with open(path) as fp:
            for line in fp:
                instance = json.loads(line)
                schema = load_yaml(f"tests/fixtures/swapi/{instance['label'].lower()}.yaml")
                break

        with VertexLinkWriter(schema) as mgr:
            output_path = path.replace(".ndjson", "-links.ndjson")
            with open(output_path, "w") as out_fp:
                with open(path) as fp:
                    for line in fp:
                        instance = mgr.insert_links(json.loads(line))
                        out_fp.write(json.dumps(instance) + "\n")
            assert os.path.exists(output_path), f"{output_path} does not exist"



def test_quantities(distribution_schema: dict):
    """Ensure that Quantities are rendered"""
    observation = distribution_schema['observation.yaml']
    expected_properties = ['valueQuantity', 'valueQuantity_unit', 'valueQuantity_value']
    for expected_property in expected_properties:
        assert expected_property in observation['properties'], f"Should have {expected_property}"
    assert observation['properties']['valueQuantity_value']['type'] == 'number'



import gzip
import importlib
import io
import pathlib
from dataclasses import dataclass
from typing import Dict, Iterator, TextIO

import click
import orjson
from fhir.resources.fhirresourcemodel import FHIRResourceModel
from pydantic import ValidationError
import logging

FHIR_CLASSES = importlib.import_module('fhir.resources')

logger = logging.getLogger(__name__)


class NaturalOrderGroup(click.Group):
    """See https://github.com/pallets/click/issues/513."""
    def list_commands(self, ctx):
        return self.commands.keys()


@dataclass
class ParseResult:
    """Results of FHIR validation of dict."""
    object: dict
    """The "raw" dictionary loaded from json string."""
    resource: FHIRResourceModel
    """If valid, the FHIR resource."""
    exception: Exception
    """If invalid, the exception."""
    path: pathlib.Path = None
    """Source file, if available."""
    offset: int = None
    """Base 0 offset of line number(ndjson) or entry(bundle)."""
    resource_id: str = None
    """Resource id of resource"""


def parse_obj(obj: Dict, validate=True, parse=True) -> ParseResult:
    """Load a dictionary into a FHIR model """
    try:
        if parse:
            assert 'resourceType' in obj, "Dict missing `resourceType`, is it a FHIR dict?"
            klass = FHIR_CLASSES.get_fhir_model_class(obj['resourceType'])
            _ = klass.parse_obj(obj)
        if validate:
            # trigger object traversal, see monkey patch below, at bottom of file
            _.dict()
        return ParseResult(object=obj, resource=_, exception=None, path=None, resource_id=_.id)
    except (ValidationError, AssertionError) as e:
        return ParseResult(object=obj, resource=None, exception=e, path=None, resource_id=obj.get('id', None))


def _is_ndjson(file_path: pathlib.Path) -> bool:
    """Open file, read all lines as json."""
    fp = _to_file(file_path)
    try:
        with fp:
            for line in fp.readlines():
                orjson.loads(line)
                break
        return True
    except Exception as e:  # noqa
        return False


def _to_file(file_path):
    """Open file appropriately."""
    if file_path.name.endswith('gz'):
        fp = io.TextIOWrapper(io.BufferedReader(gzip.GzipFile(file_path)))  # noqa
    else:
        fp = open(file_path, "rb")
    return fp


def _is_json_file(name: str) -> bool:
    """Files we are interested in"""
    if name.endswith('json.gz'):
        return True
    if name.endswith('json'):
        return True
    return False


def _has_entries(_: ParseResult):
    """"""
    if _.resource is None:
        return False
    return _.resource.resource_type in ["Bundle", "List"] and _.resource.entry is not None


def _entry_iterator(parse_result: ParseResult) -> Iterator[ParseResult]:
    """See if there are entries"""
    if not _has_entries(parse_result):
        yield parse_result
    else:
        _path = parse_result.path
        offset = 0
        if parse_result.resource.entry and len(parse_result.resource.entry) > 0:
            for _ in parse_result.resource.entry:
                if _ is None:
                    break
                if hasattr(_, 'resource'):  # BundleEntry
                    yield ParseResult(object=None, path=_path, resource=_.resource, offset=offset, exception=None)
                elif hasattr(_, 'item'):  # ListEntry
                    yield ParseResult(object=None, path=_path, resource=_.item, offset=offset, exception=None)
                else:
                    yield ParseResult(object=None, path=_path, resource=_.item, offset=offset, exception=None)
                offset += 1
    pass


def directory_reader(
        directory_path: pathlib.Path,
        pattern: str = '*.*',
        parse=True,
        validate=True,
        ignore_path:str =None) -> Iterator[ParseResult]:
    """Extract FHIR resources from directory

    Args:
        directory_path (pathlib.Path): directory to read
        pattern (str, optional): glob pattern. Defaults to '*.*'.
        parse (bool, optional): parse FHIR resources. Defaults to True.
        validate (bool, optional): validate FHIR resources. Defaults to True.
        ignore_path (str, optional): ignore files with this string in path. Defaults to None.
    """

    assert directory_path.is_dir(), f"{directory_path.name} is not a directory"

    input_files = [_ for _ in directory_path.glob(pattern) if _is_json_file(_.name)]
    for input_file in input_files:
        if ignore_path is not None and ignore_path in str(input_file):
            continue
        logger.info(input_file)
        if not input_file.is_file():
            continue
        is_ndjson = _is_ndjson(input_file)
        fp = _to_file(input_file)
        with fp:
            if is_ndjson:
                offset = 0
                for line in fp.readlines():
                    parse_result = parse_obj(orjson.loads(line), validate=validate, parse=parse)
                    parse_result.path = input_file
                    parse_result.offset = offset
                    for _ in _entry_iterator(parse_result):
                        # print(_.offset, 'is_ndjson', input_file, parse_result.resource.id, parse_result.resource.resource_type)
                        yield _
                    offset += 1
            else:
                # look for json bundles
                _ = orjson.loads(fp.read())
                # not a bundle
                parse_result = parse_obj(_, validate)
                parse_result.path = input_file
                parse_result.offset = 0
                for _ in _entry_iterator(parse_result):
                    yield _
                continue


class EmitterContextManager:
    """Maintain file pointers to output directory."""

    def __init__(self, output_path: str, verbose=False, file_mode="w"):
        """Ensure output_path exists, init emitter dict."""
        output_path = pathlib.Path(output_path)
        if not output_path.exists():
            output_path.mkdir(parents=True)
        assert output_path.is_dir(), f"{output_path} not a directory?"

        self.output_path = output_path
        """destination directory"""
        self.emitters = {}
        """open file pointers"""
        self.verbose = verbose
        """log activity"""
        self.file_mode = file_mode
        """mode for file opens"""

    def __enter__(self):
        """Ensure output_path exists, init emitter dict.
        """
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Close all open files."""
        for _ in self.emitters.values():
            _.close()
            if self.verbose:
                logger.info(f"wrote {_.name}")

    def emit(self, name: str) -> TextIO:
        """Maintain a hash of open files."""
        if name not in self.emitters:
            self.emitters[name] = open(self.output_path / f"{name}.ndjson", self.file_mode)
            if self.verbose:
                logger.debug(f"opened {self.emitters[name].name}")
        return self.emitters[name]



import fastjsonschema


def test_schema(distribution_schema):
    """Ensure the schema compiles as a legal json schema."""
    compiled_schema = fastjsonschema.compile(distribution_schema)
    assert compiled_schema



import json
import pytest

import os

from dictionaryutils import DataDictionary


@pytest.fixture()
def simplified_dir():
    """cd to simplified dir"""
    # setup
    cwd = os.getcwd()
    os.chdir('iceberg/schemas/simplified')
    # run test
    yield 'iceberg/schemas/simplified'
    # teardown
    os.chdir(cwd)


@pytest.fixture()
def data_dictionary_from_yaml(simplified_dir) -> DataDictionary:
    """Load the schema from individual yaml files."""
    yield DataDictionary(root_dir=os.getcwd())


@pytest.fixture()
def distribution_schema(simplified_dir) -> dict:
    """Load the simplified-fhir schema"""
    yield json.load(open("simplified-fhir.json"))



import orjson
from fhir.resources.medicationadministration import MedicationAdministration

from iceberg_tools.data.simplifier import simplify, _render_dialect, SimplifierContextManager


def test_simplify_medication(distribution_schema):
    medication_administration = {
        "resourceType": "MedicationAdministration",
        "id": "97ac08a2-9c53-4a02-9035-04e42502a1d3",
        "meta": {"versionId": "1", "lastUpdated": "2023-01-26T14:31:45.031+00:00",
                 "source": "#l9MRe2cRjfS7IVVt"}, "status": "completed",
        "subject": {"reference": "Patient/74af22f2-a67d-4aa6-b932-30383fec846b"},
        "occurenceDateTime": "1997-10-02T16:08:23-04:00", "medication": {"concept": {
            "coding": [{"system": "http://www.nlm.nih.gov/research/umls/rxnorm", "code": "1719286",
                        "display": "10 ML Furosemide 10 MG/ML Injection"}],
            "text": "10 ML Furosemide 10 MG/ML Injection"}},
        "encounter": {"reference": "Encounter/4e970fac-3cc1-43d3-ab26-0756037a58a1"},
        "reason": [
            {"reference": {"reference": "Condition/83d3ecf0-0ce7-473c-ac31-1f4be6c5780d"}}]
    }

    medication_administration = MedicationAdministration(**medication_administration)
    with SimplifierContextManager():

        simplified, references = simplify(medication_administration, dialect='PFB')
        rendered = _render_dialect(simplified, references, 'PFB', schemas=distribution_schema)

        print(orjson.dumps(rendered, option=orjson.OPT_INDENT_2).decode())

        assert 'medication' in rendered['object'], "Should render CodeableReference codeable"
        assert len(rendered['relations']) > 0
        condition = next(iter([_ for _ in rendered['relations'] if _['dst_name'] == 'condition']), None)
        assert condition, "Should render Condition relation"



import pathlib
import yaml
from jsonschema.validators import Draft202012Validator


def test_yaml_anonymous_schemas(bmeg_dir):
    """Check individual anonymous schemas"""
    for file_name in pathlib.Path('').glob("*.yaml"):
        with open(file_name) as fp:
            Draft202012Validator.check_schema(yaml.load(fp, yaml.SafeLoader))



import orjson
from fhir.resources.task import Task

from iceberg_tools.data.simplifier import simplify, _render_dialect, SimplifierContextManager


def test_simplify_task(distribution_schema):
    task = {
        "resourceType": "Task", "id": "857f89ab-2808-5592-a3e1-315466bc6181",
        "meta": {"versionId": "1", "lastUpdated": "2023-01-26T15:03:23.029+00:00",
                 "source": "#Vpgkq7bDayJaUwY1"}, "text": {"status": "generated",
                                                          "div": "<div xmlns=\"http://www.w3.org/1999/xhtml\">Autogenerated task. Inserted to make data model research friendly.</div>"},
        "status": "completed", "intent": "order",
        "focus": {"reference": "Specimen/c43eab49-c9d3-52fb-9265-9db6461b847c"},
        "for": {"reference": "Patient/1efc46e2-8aad-46be-8713-13d17e3eda83"}, "input": [
            {"type": {"coding": [{"code": "specimen"}]},
             "valueReference": {"reference": "Specimen/c43eab49-c9d3-52fb-9265-9db6461b847c"}}], "output": [
            {"type": {"coding": [{"code": "DocumentReference"}]},
             "valueReference": {"reference": "DocumentReference/64ca4c81-d466-5eb6-b57e-62238241c0e4"}}]
    }

    task = Task(**task)
    with SimplifierContextManager():

        simplified, references = simplify(task, dialect='PFB')
        rendered = _render_dialect(simplified, references, 'PFB', schemas=distribution_schema)

        print(orjson.dumps(rendered, option=orjson.OPT_INDENT_2).decode())

        assert len(rendered['relations']) > 0
        document_reference = next(iter([_ for _ in rendered['relations'] if _['dst_name'] == 'document_reference']), None)
        assert document_reference, "Should render DocumentReference relation"
